import logging
from typing import Dict, Callable

import pandas as pd

from carball.analysis.events.bump_detection.bump_analysis import BumpAnalysis
from carball.analysis.events.boost_pad_detection.pickup_analysis import PickupAnalysis
from carball.analysis.events.hit_pressure.pressure_analysis import PressureAnalysis
from carball.analysis.events.fifty_fifty.fifty_analysis import FiftyAnalysis
from carball.analysis.events.kickoff_detection.kickoff_analysis import BaseKickoff
from carball.analysis.events.carry_detection import CarryDetection
from carball.analysis.events.hit_detection.base_hit import BaseHit
from carball.analysis.events.hit_detection.hit_analysis import SaltieHit
from carball.analysis.events.dropshot.damage import create_dropshot_damage_events
from carball.analysis.events.dropshot.ball import create_dropshot_ball_events
from carball.generated.api import game_pb2
from carball.generated.api.player_pb2 import Player
from carball.json_parser.game import Game

logger = logging.getLogger(__name__)

class EventsCreator:
    """
        Handles the creation of all events that can then be later used for stats
    """

    def __init__(self, id_creator: Callable):
        self.id_creator = id_creator

    def create_events(self, game: Game, proto_game: game_pb2.Game, player_map: Dict[str, Player],
                      data_frame: pd.DataFrame, kickoff_frames: pd.DataFrame, first_touch_frames: pd.Series,
                      calculate_intensive_events: bool = False):
        """
        Creates all of the event protos.
        :param calculate_intensive_events: Indicates if expensive calculations should run to include additional stats.
        """
        goal_frames = data_frame.game.goal_number.notnull()
        self.create_boostpad_events(proto_game, data_frame)
        self.create_hit_events(game, proto_game, player_map, data_frame, kickoff_frames, first_touch_frames)
        self.calculate_kickoff_stats(game, proto_game, player_map, data_frame, kickoff_frames, first_touch_frames)
        self.calculate_ball_carries(game, proto_game, player_map, data_frame[goal_frames])
        self.create_bumps(game, proto_game, player_map, data_frame[goal_frames])
        self.create_dropshot_events(game, proto_game, player_map)

        if calculate_intensive_events:
            self.calculate_hit_pressure(game, proto_game, data_frame)
            self.calculate_fifty_fifty(game, proto_game, data_frame)
            # TODO (j-wass): calculate bumps

    def calculate_fifty_fifty(self, game: Game, proto_game: game_pb2.Game, data_frame: pd.DataFrame):
        logger.info("Calculating 50/50s.")
        fiftyAnalysis = FiftyAnalysis(game=game, proto_game=proto_game, data_frame=data_frame)
        fiftyAnalysis.calculate_fifty_fifty_stats()

    def calculate_hit_pressure(self, game: Game, proto_game: game_pb2.Game, data_frame: pd.DataFrame):
        logger.info("Calculating hit pressure.")
        pressureAnalysis = PressureAnalysis(game=game, proto_game=proto_game, data_frame=data_frame)
        pressureAnalysis.calculate_pressure_stats()

    def calculate_kickoff_stats(self, game: Game, proto_game: game_pb2.Game, player_map: Dict[str, Player],
                                data_frame, kickoff_frames, first_touch_frames):
        logger.info("Looking for kickoffs.")
        kickoffs = BaseKickoff.get_kickoffs_from_game(game, proto_game, self.id_creator, player_map, data_frame, kickoff_frames, first_touch_frames)
        logger.info("Found %s kickoffs." % len(kickoffs.keys()))

    def create_hit_events(self, game: Game, proto_game: game_pb2.Game, player_map: Dict[str, Player],
                          data_frame: pd.DataFrame, kickoff_frames: pd.DataFrame, first_touch_frames: pd.Series):
        """
        Creates all of the events for hits
        """
        logger.info("Looking for hits.")
        hits = BaseHit.get_hits_from_game(game, proto_game, self.id_creator, data_frame, first_touch_frames)
        logger.info("Found %s hits." % len(hits))

        SaltieHit.get_saltie_hits_from_game(proto_game, hits, player_map, data_frame, kickoff_frames)
        logger.info("Analysed hits.")

        # self.stats = get_stats(self)

    def calculate_ball_carries(self, game: Game, proto_game: game_pb2.Game, player_map: Dict[str, Player],
                               data_frame: pd.DataFrame):
        logger.info("Looking for carries.")
        carry_detection = CarryDetection()
        carry_data = carry_detection.filter_frames(data_frame)

        for player in player_map:
            carry_detection.create_carry_events(carry_data, player_map[player], proto_game, data_frame)
            # find now continuous data of longer than a second.
        logger.info("Found %s carries.", len(proto_game.game_stats.ball_carries))

    def create_bumps(self, game: Game, proto_game: game_pb2.Game, player_map: Dict[str, Player],
                     data_frame: pd.DataFrame):
        logger.info("Looking for bumps.")
        bumpAnalysis = BumpAnalysis(game=game, proto_game=proto_game)
        bumpAnalysis.get_bumps_from_game(data_frame)
        logger.info("Found %s bumps.", len(proto_game.game_stats.bumps))

    def create_dropshot_events(self, game: Game, proto_game: game_pb2.Game, player_map: Dict[str, Player]):
        create_dropshot_damage_events(game, proto_game)
        create_dropshot_ball_events(game, proto_game, player_map)

    def create_boostpad_events(self, proto_game: game_pb2.Game, data_frame: pd.DataFrame):
        PickupAnalysis.add_pickups(proto_game, data_frame)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api/stats/extra_mode_stats.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ...api.stats import dropshot_pb2
from ...api.stats import rumble_pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='api/stats/extra_mode_stats.proto',
  package='api.stats',
  serialized_pb=_b('\n api/stats/extra_mode_stats.proto\x12\tapi.stats\x1a\x18\x61pi/stats/dropshot.proto\x1a\x16\x61pi/stats/rumble.proto\"\xc5\x01\n\x11\x45xtraModeGoalInfo\x12\x11\n\tpre_items\x18\x01 \x01(\x08\x12\x18\n\x10scored_with_item\x18\x02 \x01(\x08\x12%\n\tused_item\x18\x03 \x01(\x0e\x32\x12.api.stats.PowerUp\x12.\n\rdropshot_tile\x18\x04 \x01(\x0b\x32\x17.api.stats.DropshotTile\x12\x15\n\rphase_1_tiles\x18\x05 \x01(\x05\x12\x15\n\rphase_2_tiles\x18\x06 \x01(\x05\"U\n\x12\x45xtraModeBallStats\x12?\n\x14\x64ropshot_phase_stats\x18\x01 \x03(\x0b\x32!.api.stats.DropshotBallPhaseStats')
  ,
  dependencies=[dropshot_pb2.DESCRIPTOR,rumble_pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_EXTRAMODEGOALINFO = _descriptor.Descriptor(
  name='ExtraModeGoalInfo',
  full_name='api.stats.ExtraModeGoalInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pre_items', full_name='api.stats.ExtraModeGoalInfo.pre_items', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='scored_with_item', full_name='api.stats.ExtraModeGoalInfo.scored_with_item', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='used_item', full_name='api.stats.ExtraModeGoalInfo.used_item', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='dropshot_tile', full_name='api.stats.ExtraModeGoalInfo.dropshot_tile', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='phase_1_tiles', full_name='api.stats.ExtraModeGoalInfo.phase_1_tiles', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='phase_2_tiles', full_name='api.stats.ExtraModeGoalInfo.phase_2_tiles', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=98,
  serialized_end=295,
)


_EXTRAMODEBALLSTATS = _descriptor.Descriptor(
  name='ExtraModeBallStats',
  full_name='api.stats.ExtraModeBallStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='dropshot_phase_stats', full_name='api.stats.ExtraModeBallStats.dropshot_phase_stats', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=297,
  serialized_end=382,
)

_EXTRAMODEGOALINFO.fields_by_name['used_item'].enum_type = rumble_pb2._POWERUP
_EXTRAMODEGOALINFO.fields_by_name['dropshot_tile'].message_type = dropshot_pb2._DROPSHOTTILE
_EXTRAMODEBALLSTATS.fields_by_name['dropshot_phase_stats'].message_type = dropshot_pb2._DROPSHOTBALLPHASESTATS
DESCRIPTOR.message_types_by_name['ExtraModeGoalInfo'] = _EXTRAMODEGOALINFO
DESCRIPTOR.message_types_by_name['ExtraModeBallStats'] = _EXTRAMODEBALLSTATS

ExtraModeGoalInfo = _reflection.GeneratedProtocolMessageType('ExtraModeGoalInfo', (_message.Message,), dict(
  DESCRIPTOR = _EXTRAMODEGOALINFO,
  __module__ = 'api.stats.extra_mode_stats_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.ExtraModeGoalInfo)
  ))
_sym_db.RegisterMessage(ExtraModeGoalInfo)

ExtraModeBallStats = _reflection.GeneratedProtocolMessageType('ExtraModeBallStats', (_message.Message,), dict(
  DESCRIPTOR = _EXTRAMODEBALLSTATS,
  __module__ = 'api.stats.extra_mode_stats_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.ExtraModeBallStats)
  ))
_sym_db.RegisterMessage(ExtraModeBallStats)


# @@protoc_insertion_point(module_scope)



from carball.tests.utils import get_raw_replays, run_analysis_test_on_replay

from carball.analysis.analysis_manager import AnalysisManager

# Testclass for fifty-fifties (collision with ball and multiple players) and their consequential stats (ex dunks)
class Test_Fifties:
    def test_calculate_fifty_fifties(self, replay_cache):
        def test(analysis: AnalysisManager):
            proto_game = analysis.get_protobuf_data()
            fifties = proto_game.game_stats.fifty_fifties
            # Tuples are (starting frame, ending frame, # hits, # unique players)
            expected_tuples = [
            (173, 177, 4, 2),
            (259, 261, 2, 2),
            (539, 540, 2, 2),
            (1212, 1214, 2, 2),
            (1676, 1679, 2, 2),
            (2261, 2265, 3, 2),
            (2634, 2636, 2, 2),
            (3172, 3174, 3, 2),
            (4132, 4134, 2, 2),
            (4774, 4776, 2, 2),
            (5124, 5127, 2, 2),
            (5224, 5226, 2, 2),
            (6260, 6262, 2, 2),
            (7495, 7496, 2, 2),
            (7916, 7921, 3, 2),
            (8203, 8207, 2, 2),
            (8855, 8857, 3, 2),
            (8906, 8909, 2, 2),
            (9291, 9294, 3, 2),
            (10257, 10259, 2, 2),
            ]
            assert(len(fifties) == len(expected_tuples))
            for i in range(len(fifties)):
                fifty = fifties[i]
                fifty_tuple = (fifty.starting_frame, fifty.ending_frame, len(fifty.hits), len(fifty.players))
                assert(fifty_tuple == expected_tuples[i])

        # Skip test cache since this test is calculating intensive events.
        run_analysis_test_on_replay(test, get_raw_replays()["OCE_RLCS_7_CARS"], calculate_intensive_events=True)

    def test_fifties_empty_if_not_intensive(self, replay_cache):
        def test(analysis: AnalysisManager):
            proto_game = analysis.get_protobuf_data()
            fifties = proto_game.game_stats.fifty_fifties
            assert(len(fifties) == 0)

        # Skip test cache since this test is calculating intensive events.
        run_analysis_test_on_replay(test, get_raw_replays()["OCE_RLCS_7_CARS"], calculate_intensive_events=False)

    def test_dunks(self, replay_cache):
        def test(analysis: AnalysisManager):
            proto_game = analysis.get_protobuf_data()
            hits = proto_game.game_stats.hits
            expected_dunk_frames = [
            261,
            1679,
            2265,
            3174,
            4776,
            5226,
            7921,
            8207,
            ]
            dunk_frames = list(map(lambda x: x.frame_number,filter(lambda y: y.dunk,hits)))
            assert(expected_dunk_frames==dunk_frames)

        # Skip test cache since this test is calculating intensive events.
        run_analysis_test_on_replay(test, get_raw_replays()["OCE_RLCS_7_CARS"], calculate_intensive_events=True)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api/stats/rumble.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='api/stats/rumble.proto',
  package='api.stats',
  serialized_pb=_b('\n\x16\x61pi/stats/rumble.proto\x12\tapi.stats\"X\n\x0bRumbleStats\x12\x31\n\x0crumble_items\x18\x01 \x03(\x0b\x32\x1b.api.stats.RumbleItemsUsage\x12\x16\n\x0epre_item_goals\x18\x02 \x01(\x05\"h\n\x10RumbleItemsUsage\x12 \n\x04item\x18\x01 \x01(\x0e\x32\x12.api.stats.PowerUp\x12\x0c\n\x04used\x18\x02 \x01(\x05\x12\x0e\n\x06unused\x18\x03 \x01(\x05\x12\x14\n\x0c\x61verage_hold\x18\x04 \x01(\x02*\xd7\x01\n\x07PowerUp\x12\x0f\n\x0b\x42\x41LL_FREEZE\x10\x01\x12\x17\n\x13\x42\x41LL_GRAPPLING_HOOK\x10\x02\x12\x0e\n\nBALL_LASSO\x10\x03\x12\x0c\n\x08\x42\x41TARANG\x10\x03\x12\x0f\n\x0b\x42\x41LL_SPRING\x10\x04\x12\x0f\n\x0b\x42\x41LL_VELCRO\x10\x05\x12\x12\n\x0e\x42OOST_OVERRIDE\x10\x06\x12\x0e\n\nCAR_SPRING\x10\x07\x12\x10\n\x0cGRAVITY_WELL\x10\x08\x12\x0e\n\nSTRONG_HIT\x10\t\x12\x0b\n\x07SWAPPER\x10\n\x12\x0b\n\x07TORNADO\x10\x0b\x1a\x02\x10\x01')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

_POWERUP = _descriptor.EnumDescriptor(
  name='PowerUp',
  full_name='api.stats.PowerUp',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='BALL_FREEZE', index=0, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BALL_GRAPPLING_HOOK', index=1, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BALL_LASSO', index=2, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BATARANG', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BALL_SPRING', index=4, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BALL_VELCRO', index=5, number=5,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BOOST_OVERRIDE', index=6, number=6,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CAR_SPRING', index=7, number=7,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GRAVITY_WELL', index=8, number=8,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STRONG_HIT', index=9, number=9,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SWAPPER', index=10, number=10,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TORNADO', index=11, number=11,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=_descriptor._ParseOptions(descriptor_pb2.EnumOptions(), _b('\020\001')),
  serialized_start=234,
  serialized_end=449,
)
_sym_db.RegisterEnumDescriptor(_POWERUP)

PowerUp = enum_type_wrapper.EnumTypeWrapper(_POWERUP)
BALL_FREEZE = 1
BALL_GRAPPLING_HOOK = 2
BALL_LASSO = 3
BATARANG = 3
BALL_SPRING = 4
BALL_VELCRO = 5
BOOST_OVERRIDE = 6
CAR_SPRING = 7
GRAVITY_WELL = 8
STRONG_HIT = 9
SWAPPER = 10
TORNADO = 11



_RUMBLESTATS = _descriptor.Descriptor(
  name='RumbleStats',
  full_name='api.stats.RumbleStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='rumble_items', full_name='api.stats.RumbleStats.rumble_items', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='pre_item_goals', full_name='api.stats.RumbleStats.pre_item_goals', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=37,
  serialized_end=125,
)


_RUMBLEITEMSUSAGE = _descriptor.Descriptor(
  name='RumbleItemsUsage',
  full_name='api.stats.RumbleItemsUsage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='item', full_name='api.stats.RumbleItemsUsage.item', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='used', full_name='api.stats.RumbleItemsUsage.used', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='unused', full_name='api.stats.RumbleItemsUsage.unused', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='average_hold', full_name='api.stats.RumbleItemsUsage.average_hold', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=127,
  serialized_end=231,
)

_RUMBLESTATS.fields_by_name['rumble_items'].message_type = _RUMBLEITEMSUSAGE
_RUMBLEITEMSUSAGE.fields_by_name['item'].enum_type = _POWERUP
DESCRIPTOR.message_types_by_name['RumbleStats'] = _RUMBLESTATS
DESCRIPTOR.message_types_by_name['RumbleItemsUsage'] = _RUMBLEITEMSUSAGE
DESCRIPTOR.enum_types_by_name['PowerUp'] = _POWERUP

RumbleStats = _reflection.GeneratedProtocolMessageType('RumbleStats', (_message.Message,), dict(
  DESCRIPTOR = _RUMBLESTATS,
  __module__ = 'api.stats.rumble_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.RumbleStats)
  ))
_sym_db.RegisterMessage(RumbleStats)

RumbleItemsUsage = _reflection.GeneratedProtocolMessageType('RumbleItemsUsage', (_message.Message,), dict(
  DESCRIPTOR = _RUMBLEITEMSUSAGE,
  __module__ = 'api.stats.rumble_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.RumbleItemsUsage)
  ))
_sym_db.RegisterMessage(RumbleItemsUsage)


_POWERUP.has_options = True
_POWERUP._options = _descriptor._ParseOptions(descriptor_pb2.EnumOptions(), _b('\020\001'))
# @@protoc_insertion_point(module_scope)



import json
import numpy as np


class CarballJsonEncoder(json.JSONEncoder):
    def default(self, o):
        # it cannot normally serialize np.int64 and possibly other np data types
        if type(o).__module__ == np.__name__:
            return o.item()
        return super(CarballJsonEncoder, self).default(o)




from carball.analysis.analysis_manager import AnalysisManager

from carball.tests.utils import run_analysis_test_on_replay, get_specific_replays


class DBTest():
    def test_offline_replay(self, replay_cache):

        def test(analysis: AnalysisManager):
            assert analysis.get_protobuf_data() is not None
            assert not analysis.get_protobuf_data().game_metadata.is_invalid_analysis
            game = analysis.get_protobuf_data()
            assert game.game_metadata.match_guid, game.game_metadata.id

        run_analysis_test_on_replay(test, replay_list=get_specific_replays()["OFFLINE"], cache=replay_cache)



import os
import platform
from subprocess import call


def is_windows():
    return platform.system() == 'Windows'


current_dir = os.path.dirname(os.path.dirname(__file__))


proto_dir = os.path.join(current_dir, 'carball', 'generated')


def get_proto():
    if os.getenv('PROTOC_PATH'):
        return os.getenv('PROTOC_PATH')


    if is_windows():
        return os.path.join(proto_dir, 'protoc.exe')
    else:
        import shutil
        result = shutil.which('protoc')
        if result is not None:
            return result
        else:
            return os.path.join(proto_dir, 'binaries', 'protoc')


def split_to_list(drive_and_path):
    path = os.path.splitdrive(drive_and_path)[1]
    folders = []
    while 1:
        path, folder = os.path.split(path)

        if folder != "":
            folders.append(folder)
        else:
            if path != "":
                folders.append(path)

            break

    folders.reverse()
    return folders


def get_dir():
    return current_dir


def get_deepness(top_level_dir, path_list):
    return len(path_list) - path_list.index(top_level_dir)


def get_file_list(top_level_dir, exclude_dir=None, file_extension='.py'):
    proto_directories = [x[0] for x in os.walk(get_dir()) if top_level_dir in x[0] and '__pycache__' not in x[0]]

    file_result = []

    path_lists = []
    for path in proto_directories:
        if exclude_dir is not None and exclude_dir in path:
            continue
        path_list = split_to_list(path)
        deepness = get_deepness(top_level_dir, path_list)
        left_over_paths = path_list[-deepness:]
        path_lists.append((path, deepness, left_over_paths))
    for path_item in path_lists:
        path = path_item[0]
        only_files = [(os.path.join(path, f), f) for f in os.listdir(path) if os.path.isfile(os.path.join(path, f))
                      and file_extension in f and '__init__' not in f]
        for file in only_files:
            file_result.append((path_item[1], file[0]))
    return file_result


def create_proto_files():
    print('###CREATING PROTO FILES###')
    file_list = get_file_list(top_level_dir='api', file_extension='.proto')
    for file in file_list:
        path = file[0]
        file = file[1]
        print('creating proto file', file, end='\t')
        result = call([get_proto(), '--python_out=' + proto_dir, '--proto_path=' + current_dir, file])
        if result != 0:
            raise ValueError(result)
        print(result)


if __name__ == "__main__":
    create_proto_files()



import logging
from .base import *

logger = logging.getLogger(__name__)


class PlayerHandler(BaseActorHandler):
    type_name = 'TAGame.Default__PRI_TA'

    def update(self, actor: dict, frame_number: int, time: float, delta: float) -> None:
        if 'Engine.PlayerReplicationInfo:PlayerName' not in actor:
            return

        actor_id = actor['Id']
        player_dict = {
            'name': actor["Engine.PlayerReplicationInfo:PlayerName"],
        }
        # Conditionally add ['team'] key to player_dict
        player_team = actor.get("Engine.PlayerReplicationInfo:Team", {}).get('actor', None)
        if player_team is not None and player_team != -1:
            player_dict['team'] = player_team

        if "TAGame.PRI_TA:PartyLeader" in actor and actor["TAGame.PRI_TA:PartyLeader"] is not None:
            try:
                actor_type = list(actor["Engine.PlayerReplicationInfo:UniqueId"]['remote_id'].keys())[0]

                # handle UniqueID for plays_station and switch
                unique_id = None
                if actor_type == "PlayStation" or actor_type == "PsyNet":
                    actor_name = actor["Engine.PlayerReplicationInfo:PlayerName"]
                    for player_stat in self.parser.game.properties['PlayerStats']:
                        if actor_name == player_stat['Name']:
                            unique_id = str(player_stat['OnlineID'])
                if unique_id is None:
                    unique_id = str(actor['Engine.PlayerReplicationInfo:UniqueId']['remote_id'][actor_type])

                # only process if party_leader id exists
                if "remote_id" in actor["TAGame.PRI_TA:PartyLeader"]:
                    leader_actor_type = list(actor["TAGame.PRI_TA:PartyLeader"]["remote_id"].keys())[0]
                    leader = None
                    if leader_actor_type == "PlayStation" or leader_actor_type == "PsyNet":
                        leader_name = actor["TAGame.PRI_TA:PartyLeader"]["remote_id"][leader_actor_type]['name']

                        for player_stat in self.parser.game.properties['PlayerStats']:
                            if leader_name == player_stat['Name']:
                                leader = str(player_stat['OnlineID'])

                    if leader is None:  # leader is not using play_station nor switch (ie. xbox or steam)
                        leader = str(actor["TAGame.PRI_TA:PartyLeader"]["remote_id"][leader_actor_type])

                    if leader in self.parser.parties:
                        if unique_id not in self.parser.parties[leader]:
                            self.parser.parties[leader].append(unique_id)
                    else:
                        self.parser.parties[leader] = [unique_id]

            except KeyError:
                logger.warning('Could not get party leader for actor id: ' + str(actor_id))

        if actor_id not in self.parser.player_dicts:
            # add new player
            self.parser.player_dicts[actor_id] = player_dict

            logger.debug('Found player actor: %s (id: %s)' % (player_dict['name'], actor_id))
            self.parser.player_data[actor_id] = {}

        self.parser.player_data[actor_id][frame_number] = {}

        # update player_dicts
        for _k, _v in {**actor, **player_dict}.items():
            self.parser.player_dicts[actor_id][_k] = _v

        if delta != 0:
            self.parser.player_data[actor_id][frame_number]['ping'] = \
                actor.get("Engine.PlayerReplicationInfo:Ping", None)
            if 'TAGame.PRI_TA:CameraSettings' in actor:
                # oldstyle camera settings
                if actor_id not in self.parser.cameras_data:
                    self.parser.cameras_data[actor_id] = actor['TAGame.PRI_TA:CameraSettings']
                ball_cam = actor.get('TAGame.CameraSettingsActor_TA:bUsingSecondaryCamera', None)
                self.parser.player_data[actor_id][frame_number]['ball_cam'] = ball_cam

            if 'TAGame.PRI_TA:TimeTillItem' in actor:
                time_till_item = actor['TAGame.PRI_TA:TimeTillItem']
                self.parser.player_data[actor_id][frame_number]['time_till_power_up'] = time_till_item



import math
from typing import Dict, List

import pandas as pd

from carball.generated.api import game_pb2
from carball.generated.api.player_pb2 import Player
from carball.generated.api.stats.player_stats_pb2 import PlayerStats
from carball.generated.api.stats.team_stats_pb2 import TeamStats
from carball.generated.api.stats.events_pb2 import RumbleItemEvent
from carball.generated.api.stats.rumble_pb2 import PowerUp, RumbleStats
from carball.json_parser.game import Game
from carball.analysis.stats.stats import BaseStat
from carball.generated.api.metadata.game_metadata_pb2 import RANKED_RUMBLE, UNRANKED_RUMBLE


class RumbleItemStat(BaseStat):

    def calculate_player_stat(self, player_stat_map: Dict[str, PlayerStats], game: Game, proto_game: game_pb2.Game,
                              player_map: Dict[str, Player], data_frame: pd.DataFrame):
        if not is_rumble_enabled(game):
            return

        for player_key, stats in player_stat_map.items():
            player_name = player_map[player_key].name
            player_data_frame = data_frame[player_name]

            events = _get_power_up_events(player_map[player_key], player_data_frame, game,
                                          proto_game.game_stats.rumble_items)

            _calculate_rumble_stats(stats.rumble_stats, events, data_frame['game'])

    def calculate_team_stat(self, team_stat_list: Dict[int, TeamStats], game: Game, proto_game: game_pb2.Game,
                            player_map: Dict[str, Player], data_frame: pd.DataFrame):
        if not is_rumble_enabled(game):
            return

        orange_ids = list(map(lambda x: x.id.id, filter(lambda x: x.is_orange, player_map.values())))
        blue_ids = list(map(lambda x: x.id.id, filter(lambda x: not x.is_orange, player_map.values())))

        orange_events = list(filter(lambda x: x.player_id.id in orange_ids, proto_game.game_stats.rumble_items))
        blue_events = list(filter(lambda x: x.player_id.id in blue_ids, proto_game.game_stats.rumble_items))

        _calculate_rumble_stats(team_stat_list[1].rumble_stats, orange_events, data_frame['game'])
        _calculate_rumble_stats(team_stat_list[0].rumble_stats, blue_events, data_frame['game'])


def is_rumble_enabled(game: Game) -> bool:
    """
    Check whether rumble is enabled or not.

    :param game: parsed game object
    :return: True if rumble
    """
    if game is None or game.game_info is None:
        return False
    return game.game_info.playlist in [RANKED_RUMBLE, UNRANKED_RUMBLE] or \
           (game.game_info.rumble_mutator is not None and
            game.game_info.rumble_mutator.startswith('Archetypes.Mutators.SubRules.ItemsMode'))


def _get_power_up_events(player: Player, df: pd.DataFrame, game: Game, proto_rumble_item_events) \
        -> List[RumbleItemEvent]:
    """
    Finds the item get and item use events

    :param player: Player info protobuf. Get's the id from here
    :param df: player dataframe, assumes the frames between goal and kickoff are already discarded
    :param game: game object
    :param proto_rumble_item_events: protobuf repeated api.stats.RumbleItemEvent
    :return list of rumble events
    """
    events = []
    if 'power_up_active' in df and 'power_up' in df:
        if 'time_till_power_up' not in df:
            # someone actually uploaded a 10 second replay of the end of the match..
            df['time_till_power_up'] = math.nan
        df = df[['time_till_power_up', 'power_up', 'power_up_active']]

        ranges = [(game.kickoff_frames[i], game.kickoff_frames[i + 1]) for i in range(len(game.kickoff_frames) - 1)]
        ranges.append((game.kickoff_frames[-1], df.index[-1]))
        data_frames = map(lambda x: df.loc[x[0]:x[1] - 1], ranges)
        data_frames = map(_squash_power_up_df, data_frames)

        for data_frame in data_frames:

            if len(data_frame) == 0:
                # goal before items
                continue

            if not math.isnan(data_frame.iloc[0]['power_up_active']):
                # happens when kickoff starts with power ups after a goal that was scored less then 1 second before
                # time's up
                data_frame.loc[-1] = [0.0, math.nan, math.nan]
                data_frame.sort_index(inplace=True)

            prev_row = data_frame.iloc[0]
            proto_current_item = None
            demoed = False

            for i, row in data_frame.iloc[1:].iterrows():
                if math.isnan(prev_row['power_up_active']):
                    if row['power_up_active'] == False:
                        if not demoed:
                            # Rumble item get event
                            proto_current_item = proto_rumble_item_events.add()
                            proto_current_item.frame_number_get = i
                            proto_current_item.item = PowerUp.Value(row['power_up'].upper())
                            proto_current_item.player_id.id = player.id.id
                        else:
                            # back from the dead
                            demoed = False

                    if row['power_up_active'] == True:
                        # immediately used items (mostly bots)
                        tmp_item = proto_rumble_item_events.add()
                        tmp_item.frame_number_get = i
                        tmp_item.frame_number_use = i
                        tmp_item.item = PowerUp.Value(row['power_up'].upper())
                        tmp_item.player_id.id = player.id.id
                        events.append(tmp_item)

                elif prev_row['power_up_active'] == False:
                    if row['power_up_active'] == True or \
                            math.isnan(row['power_up_active']) and prev_row['power_up'] == 'ball_freeze':
                        # Rumble item use event
                        # When a spiked ball is frozen, there is not 'ball_freeze,True' row, it just gets deleted
                        # immediately
                        # Could also happen when the freeze is immediately broken
                        # in theory this should not happen with other power ups?
                        proto_current_item.frame_number_use = i
                        events.append(proto_current_item)
                        proto_current_item = None
                    elif math.isnan(row['power_up_active']):
                        # happens when player is demoed
                        demoed = True

                prev_row = row

            if proto_current_item is not None:
                # unused item
                events.append(proto_current_item)
                proto_current_item = None

    return events


def _squash_power_up_df(df: pd.DataFrame):
    """
    Remove all the rows with repeated 'power_up_active'. The frames are kept whenever the value is changed.
    """
    a = df['power_up_active']
    a = a.loc[(a.shift(-1).isnull() ^ a.isnull()) | (a.shift(1).isnull() ^ a.isnull()) | ~a.isnull()]
    a = a.loc[(a.shift(1) != a) | (a.shift(-1) != a)]

    # Drop any false values that come after true values
    while len(a.loc[(a.shift(1) == True) & (a == False)]) > 0:
        a = a.loc[(a.shift(1) != True) | (a != False)]

    df = pd.concat([df[['time_till_power_up', 'power_up']], a], axis=1, join='inner')

    return df


def _calculate_rumble_stats(rumble_proto: RumbleStats, events: List[RumbleItemEvent], game_df: pd.DataFrame):
    """
    Calculate rumble stats for all items

    :param rumble_proto: proto for stats
    :param events: list of rumble events
    :param game_df: game dataframe, used for getting the time delta
    """
    for power_up in set(PowerUp.values()):
        item_stats = _calculate_rumble_stats_for_power_up(events, power_up, game_df)

        rumble_item_proto = rumble_proto.rumble_items.add()
        rumble_item_proto.item = power_up
        rumble_item_proto.used = item_stats['used']
        rumble_item_proto.unused = item_stats['unused']
        rumble_item_proto.average_hold = item_stats['average_hold']


def _calculate_rumble_stats_for_power_up(events: List[RumbleItemEvent], power_up: int, game_df: pd.DataFrame) -> Dict:
    """
    Calculate rumble statistics (used, unused, average hold) based on rumble events.

    :param events: list of rumble events
    :param power_up: power up enum
    :param game_df: game dataframe, used for getting the time delta
    :return: stats
    """
    item_events = filter(lambda x: x.item == power_up, events)
    base = {'used': 0, 'unused': 0, 'average_hold': 0}

    for event in item_events:
        if event.frame_number_use != -1:
            base['used'] += 1
            base['average_hold'] += game_df.loc[event.frame_number_use]['time'] - \
                                    game_df.loc[event.frame_number_get]['time']
        else:
            base['unused'] += 1

    if base['used'] > 0:
        base['average_hold'] /= base['used']

    return base



import logging
from typing import Dict, Callable

import numpy
import pandas as pd

from carball.analysis.stats.utils.pandas_utils import sum_deltas_start_end_frame
from carball.generated.api import game_pb2
from carball.generated.api.metadata.game_metadata_pb2 import Goal
from carball.generated.api.player_pb2 import Player
from carball.generated.api.stats.kickoff_pb2 import KickoffStats
from carball.generated.api.stats import kickoff_pb2 as kickoff
from carball.json_parser.game import Game


logger = logging.getLogger(__name__)


class BaseKickoff:

    @staticmethod
    def get_kickoffs_from_game(game: Game, proto_game: game_pb2, id_creation:Callable,
                               player_map: Dict[str, Player],
                               data_frame: pd.DataFrame, kickoff_frames: pd.DataFrame,
                               first_touch_frames: pd.DataFrame) -> Dict[int, KickoffStats]:
        kickoffs = dict()
        goals = proto_game.game_metadata.goals
        num_goals = len(goals)
        last_frame = data_frame.last_valid_index()
        first_frame = data_frame.first_valid_index()
        for index, frame in enumerate(kickoff_frames):
            starting_kickoff_time = data_frame.game.time[frame]
            cur_kickoff = proto_game.game_stats.kickoff_stats.add()
            end_frame = first_touch_frames[index]
            smaller_data_frame = data_frame.loc[max(first_frame, frame - 1):  min(end_frame + 20, last_frame)]
            cur_kickoff.start_frame = frame
            cur_kickoff.touch_frame = end_frame
            ending_time = smaller_data_frame['game']['time'][end_frame]
            time = cur_kickoff.touch_time = ending_time - starting_kickoff_time
            differs = smaller_data_frame['game']['time'][frame:end_frame].diff()
            summed_time_diff = differs.sum()
            summed_time = smaller_data_frame['game']['delta'][frame:end_frame].sum()
            if summed_time > 0:
                cur_kickoff.touch_time = summed_time
            logger.error("STRAIGHT TIME " + str(time))
            logger.error("SUM TIME" + str(summed_time))
            sum_vs_adding_diff = time - summed_time


            # find who touched the ball first
            closest_player_distance = 10000000
            closest_player_id = 0

            if index < num_goals:
                BaseKickoff.get_goal_data(cur_kickoff, goals[index], data_frame)

            # get player stats
            for player in player_map.values():
                if player.name not in data_frame:
                    continue
                kickoff_player = BaseKickoff.get_player_stats(cur_kickoff, player, smaller_data_frame, frame, end_frame)

                if kickoff_player.ball_dist < closest_player_distance:
                    closest_player_distance = kickoff_player.ball_dist
                    closest_player_id = player.id.id

            if closest_player_distance != 10000000:
                # Todo use hit analysis
                cur_kickoff.touch.first_touch_player.id = closest_player_id
            cur_kickoff.type = BaseKickoff.get_kickoff_type(cur_kickoff.touch.players)
            kickoffs[frame] = cur_kickoff
        return kickoffs

    @staticmethod
    def get_player_stats(cur_kickoff, player, data_frame: pd.DataFrame, start_frame: int, end_frame: int):

        kickoff_player = cur_kickoff.touch.players.add()
        kickoff_player.player.id = player.id.id

        kickoff_player.kickoff_position, kickoff_player.start_left = BaseKickoff.get_kickoff_position(player,
                                                                                                      data_frame,
                                                                                                      start_frame)

        kickoff_player.touch_position = BaseKickoff.get_touch_position(player, data_frame, start_frame, end_frame)
        kickoff_player.boost = data_frame[player.name]['boost'][end_frame]
        kickoff_player.ball_dist = BaseKickoff.get_dist(data_frame, player.name, end_frame)
        kickoff_player.player_position.pos_x = data_frame[player.name]['pos_x'][end_frame]
        kickoff_player.player_position.pos_y = data_frame[player.name]['pos_y'][end_frame]
        kickoff_player.player_position.pos_z = data_frame[player.name]['pos_z'][end_frame]

        kickoff_player.start_position.pos_x = data_frame[player.name]['pos_x'][start_frame]
        kickoff_player.start_position.pos_y = data_frame[player.name]['pos_y'][start_frame]
        kickoff_player.start_position.pos_z = data_frame[player.name]['pos_z'][start_frame]
        BaseKickoff.set_jumps(kickoff_player, player, data_frame, start_frame)
        return kickoff_player

    @staticmethod
    def set_jumps(kPlayer, player, data_frame, frame):
        jump_active_df        = data_frame[player.name]['jump_active']
        if 'boost_collect' in data_frame[player.name].keys():
            collected_boost_df    = data_frame[player.name]['boost_collect']
            collected_boost_df = collected_boost_df[collected_boost_df > 34]
            if len(collected_boost_df) > 0:
                kPlayer.boost_time = data_frame['game']['delta'][frame:collected_boost_df.index.values[0]].sum()

        # Make sure that we are not doing diffs on booleans
        jump_active_df = jump_active_df.astype(float)

        # check the kickoff frames (and then some) for jumps & big boost collection
        jump_active_df = jump_active_df[jump_active_df.diff(1) > 0]
        BaseKickoff.add_jumps(kPlayer, data_frame, frame, jump_active_df)

        """for f in range(frame, )):
            if boost:
                if collected_boost_df[f] == True:
                    
            if jump_active_df[f] != jump_active_df[f-1] or double_jump_active_df[f] != double_jump_active_df[f-1]:
                kPlayer.jumps.append(data_frame['game']['delta'][frame:f].sum())"""

    @staticmethod
    def add_jumps(kPlayer, data_frame, frame, jumps):
        pass

    @staticmethod
    def get_kickoff_type(players: list):
        #
        diagonals = [player.kickoff_position for player in players].count(0)
        offcenter = [player.kickoff_position for player in players].count(1)
        goalies   = [player.kickoff_position for player in players].count(2)
        if len(players) == 6:
            # 3's
            if diagonals == 4:
                if offcenter == 2:
                    return kickoff.THREES_DIAG_DIAG_OFFCENT
                if goalies == 2:
                    return kickoff.THREES_DIAG_DIAG_GOAL
            if diagonals == 2:
                if offcenter == 4:
                    return kickoff.THREES_DIAG_OFFCENT_OFFCENT
                if offcenter == 2:
                    return kickoff.THREES_DIAG_OFFCENT_GOAL
            if offcenter == 4:
                return kickoff.THREES_OFFCENT_OFFCENT_GOAL
        if len(players) == 4:
            if diagonals == 4:
                return kickoff.TWOS_DIAG_DIAG
            if diagonals == 2:
                if offcenter == 2:
                    return kickoff.TWOS_DIAG_OFFCENT
                if goalies == 2:
                    return kickoff.TWOS_DIAG_GOAL
            if offcenter == 4:
                return kickoff.TWOS_OFFCENT_OFFCENT
            if offcenter == 2:
                if goalies == 2:
                    return kickoff.TWOS_OFFCENT_GOAL
        if len(players) == 2:
            if diagonals == 2:
                return kickoff.DUEL_DIAG
            if offcenter == 2:
                return kickoff.DUEL_OFFCENT
            if goalies == 2:
                return kickoff.DUEL_GOAL
        return kickoff.UNKNOWN_KICKOFF_TYPE

    @staticmethod
    def get_kickoff_position(player_class: Player, data_frame: pd.DataFrame, frame: int):
        player = player_class.name
        player_df = data_frame[player]
        pos_x = player_df['pos_x'][frame]
        start_left = not ((pos_x < 0) ^ player_class.is_orange)
        kickoff_position = kickoff.UNKNOWN_KICKOFF_POS
        if abs(abs(pos_x) - 2050) < 100:
            kickoff_position = kickoff.DIAGONAL
        elif abs(abs(pos_x) - 256) < 100:
            kickoff_position = kickoff.OFFCENTER
        elif abs(abs(pos_x)) < 4:
            kickoff_position = kickoff.GOALIE

        return kickoff_position, start_left

    @staticmethod
    def get_dist(data_frame: pd.DataFrame, player: str, frame: int):
        player_df = data_frame[player]
        dist = (player_df['pos_x'][frame]**2 + player_df['pos_y'][frame]**2 + player_df['pos_z'][frame]**2)**(0.5)
        return dist

    @staticmethod
    def get_afk(data_frame: pd.DataFrame, player: str, frame: int, kick_frame: int):
        player_df = data_frame[player]
        return (player_df['pos_x'][frame] == player_df['pos_x'][kick_frame] and
                player_df['pos_y'][frame] == player_df['pos_y'][kick_frame] and
                player_df['pos_z'][frame] == player_df['pos_z'][kick_frame])

    @staticmethod
    def get_touch_position(player: Player, data_frame: pd.DataFrame, k_frame: int, end_frame: int):
        player_df = data_frame[player.name]
        x = abs(player_df['pos_x'][end_frame])
        y = abs(player_df['pos_y'][end_frame])
        if BaseKickoff.get_dist(data_frame, player.name, end_frame) < 700:
            return kickoff.BALL
        if BaseKickoff.get_afk(data_frame, player.name, end_frame, k_frame):
            return kickoff.AFK
        if (x > 2200) and (y > 3600):
            return kickoff.BOOST
        if (x <500) and (y > 3600):
            return kickoff.GOAL
        if (x <500) and (y < 3600):
            return kickoff.CHEAT
        return kickoff.UNKNOWN_TOUCH_POS

    # Todo get what team scored next
    @classmethod
    def get_goal_data(cls, cur_kickoff: KickoffStats, current_goal: Goal, data_frame: pd.DataFrame):
        game_time = data_frame['game', 'time']
        cur_kickoff.touch.kickoff_goal = game_time[current_goal.frame_number] - game_time[cur_kickoff.touch_frame]



from typing import Dict, List

import pandas as pd

from carball.analysis.constants.basic_math import position_column_names
from carball.analysis.constants.playlist import get_team_size_from_game
from carball.analysis.stats.possession.ball_distances import BallDistanceStat
from carball.generated.api.team_pb2 import Team

from ....analysis.stats.tendencies.positional_tendencies import PositionalTendencies
from ....generated.api import game_pb2
from ....generated.api.player_pb2 import Player
from ....generated.api.stats.team_stats_pb2 import TeamStats
from ....json_parser.game import Game

# MAX_CLUMP_DISTANCE = math.sqrt(8192**2 + 10240**2) / 8
MAX_CLUMP_DISTANCE = 3278
MIN_BOONDOCKS_DISTANCE = 7000


class TeamTendencies(PositionalTendencies):

    def calculate_team_stat(self, team_stat_list: Dict[int, TeamStats], game: Game, proto_game: game_pb2.Game,
                            player_map: Dict[str, Player], data_frame: pd.DataFrame):
        for team in proto_game.teams:
            team_size = get_team_size_from_game(proto_game)
            if len(team.player_ids) <= 1 or team_size <= 1:
                # center of mass does not matter for 1s games
                continue
            player_names = [player_map[player_id.id].name for player_id in team.player_ids]
            center_of_mass = self.calculate_team_center(data_frame, player_names)
            self.get_team_tendencies(team, data_frame, center_of_mass)

            self.calculate_displacements(team, player_map, center_of_mass, data_frame, team_size)

    def calculate_team_center(self, data_frame, list_of_players) -> (pd.DataFrame, List[pd.DataFrame]):
        players = []
        for player in list_of_players:
            player_frame = data_frame[player][position_column_names]
            players.append(player_frame)

        combined = pd.concat(players)
        center_position = combined.groupby(combined.index).mean()
        return center_position

    def get_team_tendencies(self, team: Team, data_frame: pd.DataFrame, team_center: pd.DataFrame):
        self.get_tendencies(data_frame, team_center, data_frame['ball'],
                            team.is_orange, team.stats.center_of_mass.positional_tendencies,
                            self.map_player_attributes_to_predicates)

    def calculate_displacements(self, team: Team, player_map: Dict[str, Player],
                                center_of_mass: pd.DataFrame, data_frame: pd.DataFrame, team_size: int):
        player_distances_data_frame, player_distance_times, _ \
            = BallDistanceStat.calculate_player_distance_to_location(player_map, data_frame, center_of_mass)

        average_distances = []
        for player_id in team.player_ids:
            player = player_map[player_id.id]
            average_distance_from_center = player_distances_data_frame[player.id.id].mean(skipna=True)
            self.set_player_stats(player, player_distance_times, average_distance_from_center, team_size)
            player_position_with_time = pd.concat([data_frame[player.name][position_column_names],
                                 data_frame['game', 'delta'].rename('delta')], axis=1)
            if team_size > 2:
                self.set_player_positional_stats(player, center_of_mass, player_position_with_time)

            average_distances.append(average_distance_from_center)

        team.stats.center_of_mass.average_distance_from_center = sum(average_distances) / len(average_distances)

        max_distances = player_distances_data_frame.max(axis=1)
        team.stats.center_of_mass.average_max_distance_from_center = max_distances.mean(skipna=True)

        max_distances_with_delta = pd.concat([player_distances_data_frame,
                                              data_frame['game', 'delta'].rename('delta')], axis=1)

        if len(team.player_ids) == 2:
            clump_distance = MAX_CLUMP_DISTANCE / 2
        else:
            clump_distance = MAX_CLUMP_DISTANCE

        close_frames = max_distances < clump_distance
        time_clumped = max_distances_with_delta[close_frames]['delta'].sum()
        team.stats.center_of_mass.time_clumped = time_clumped

        if len(team.player_ids) == 2:
            boondocks_distance = MIN_BOONDOCKS_DISTANCE / 2
        else:
            boondocks_distance = MIN_BOONDOCKS_DISTANCE

        boondocks_frames = max_distances > boondocks_distance
        time_boondocks = max_distances_with_delta[boondocks_frames]['delta'].sum()
        team.stats.center_of_mass.time_boondocks = time_boondocks

    def set_player_stats(self, player, player_distance_times, average_distance_from_center, team_size):
        player.stats.averages.average_distance_from_center = average_distance_from_center
        player_id = player.id.id
        distance = player.stats.distance

        if team_size > 2:
            try:
                distance.time_closest_to_team_center = player_distance_times['closest_player'][player_id]
                distance.time_furthest_from_team_center = player_distance_times['furthest_player'][player_id]
            except (AttributeError, KeyError) as e:
                distance.time_closest_to_team_center = 0
                distance.time_furthest_from_team_center = 0

    def get_flipped_dataframes(self, data_frames: Dict[str, pd.DataFrame]) -> Dict[str, pd.DataFrame]:
        new_dataframes = {
            _k: _v.copy() for _k, _v in data_frames.items()
        }

        for data_frame in new_dataframes.values():
            data_frame.pos_y *= -1

        return new_dataframes

    def set_player_positional_stats(self, player, center_of_mass, distances_with_time):
        y_position = pd.concat([distances_with_time['pos_y'].rename('car_y'),
                                center_of_mass['pos_y'].rename('center_y'),
                                distances_with_time['delta']], axis=1).dropna()
        time_greater_than = y_position[y_position['car_y'] > y_position['center_y']]['delta'].sum()
        time_less_than = y_position[y_position['car_y'] < y_position['center_y']]['delta'].sum()
        if player.is_orange:
            player.stats.relative_positioning.time_behind_center_of_mass = time_greater_than
            player.stats.relative_positioning.time_in_front_of_center_of_mass = time_less_than
        else:
            player.stats.relative_positioning.time_behind_center_of_mass = time_less_than
            player.stats.relative_positioning.time_in_front_of_center_of_mass = time_greater_than



from carball.analysis.analysis_manager import AnalysisManager

from carball.tests.utils import run_analysis_test_on_replay, get_specific_replays, get_specific_answers, get_raw_replays


class Test_Hits():

    def test_num_hits_detected(self, replay_cache):
        def test(analysis: AnalysisManager, answer):
            proto_game = analysis.get_protobuf_data()
            hits = proto_game.game_stats.hits
            assert(len(hits) == answer)
            print(analysis)

        run_analysis_test_on_replay(test, replay_list=get_specific_replays()["HITS"], answers=get_specific_answers()["HITS"],
                                    cache=replay_cache)

    def test_num_kickoff_hits_detected(self, replay_cache):
        def test(analysis: AnalysisManager):
            proto_game = analysis.get_protobuf_data()
            hits = proto_game.game_stats.hits
            kickoff_counter = 0
            for hit in hits:
                if hit.is_kickoff:
                    kickoff_counter += 1
            print(analysis)

            assert(kickoff_counter == 3)

        run_analysis_test_on_replay(test, replay_list=get_raw_replays()["3_KICKOFFS"],
                                    cache=replay_cache)

    def test_num_shots_detected(self, replay_cache):
        def test(analysis: AnalysisManager, answer):
            proto_game = analysis.get_protobuf_data()
            hits = proto_game.game_stats.hits
            shot_counter = 0
            for hit in hits:
                if hit.shot:
                    shot_counter += 1

            assert(shot_counter == answer)

        run_analysis_test_on_replay(test, replay_list=get_specific_replays()["SHOTS"],
                                    answers=get_specific_answers()["SHOTS"],
                                    cache=replay_cache)

    def test_num_passes_detected(self, replay_cache):
        def test(analysis: AnalysisManager, answer):
            proto_game = analysis.get_protobuf_data()
            hits = proto_game.game_stats.hits
            pass_counter = 0
            for hit in hits:
                if hit.pass_:
                    pass_counter += 1

            assert(pass_counter == answer)

        run_analysis_test_on_replay(test, replay_list=get_specific_replays()["PASSES"],
                                    answers=get_specific_answers()["PASSES"],
                                    cache=replay_cache)

    def test_num_saves_detected(self, replay_cache):
        def test(analysis: AnalysisManager, answer):
            proto_game = analysis.get_protobuf_data()
            hits = proto_game.game_stats.hits
            counter = 0
            for hit in hits:
                if hit.save:
                    counter += 1

            assert(counter == answer)

        run_analysis_test_on_replay(test, replay_list=get_specific_replays()["SAVES"],
                                    answers=get_specific_answers()["SAVES"],
                                    cache=replay_cache)

    def test_num_aerials_detected(self, replay_cache):
        def test(analysis: AnalysisManager, answer):
            proto_game = analysis.get_protobuf_data()
            hits = proto_game.game_stats.hits
            aerial_counter = 0
            for hit in hits:
                if hit.aerial:
                    aerial_counter += 1

            assert(aerial_counter == answer)

        run_analysis_test_on_replay(test, replay_list=get_specific_replays()["AERIALS"],
                                    answers=get_specific_answers()["AERIALS"],
                                    cache=replay_cache)

    def test_num_clears_detected(self, replay_cache):
        def test(analysis: AnalysisManager, answer):
            proto_game = analysis.get_protobuf_data()
            hits = proto_game.game_stats.hits
            clear_counter = 0
            for hit in hits:
                if hit.clear:
                    clear_counter += 1

            assert(clear_counter == answer)

        run_analysis_test_on_replay(test, replay_list=get_specific_replays()["CLEARS"],
                                    answers=get_specific_answers()["CLEARS"],
                                    cache=replay_cache)

    def test_total_clears_detected(self, replay_cache):
        def test(analysis: AnalysisManager, answer):
            clears_lookup = {'76561198204422936':6, # Decka
                       '76561198050413646': 1,      # Requeim
                       '76561198058420486': 6,      # Delusion
                       '76561197998103705': 4,      # CJCJ
                       '76561198065500375': 1,      # Express
                       '76561198173645057': 2}      # Shadey
            proto_game = analysis.get_protobuf_data()
            for pl in proto_game.players:
                expected_total_clears = clears_lookup[pl.id.id]
                calculated_total_clears = pl.stats.hit_counts.total_clears
                assert(expected_total_clears == calculated_total_clears)

        run_analysis_test_on_replay(test, replay_list=get_raw_replays()["OCE_RLCS_7_CARS"],
                                    answers=get_specific_answers()["CLEARS"],
                                    cache=replay_cache)

    def test_hit_pressure(self, replay_cache):
        def test(analysis: AnalysisManager, answer):
            proto_game = analysis.get_protobuf_data()
            hits = proto_game.game_stats.hits
            expected_pressures = [100, 100, 100, 100, 0, 93, 100, 100, 94, 0, 0,
             95, 100, 100, 0, 0, 0, 0, 90, 0, 0, 94, 100, 100, 0, 99, 0, 0, 0,
             0, 0, 0, 0, 0, 100, 100, 99, 100, 0, 0, 0, 0, 100, 100, 100, 0,
             100, 0, 0, 0, 0, 100, 100, 100, 100, 0, 100, 0, 94, 0, 96, 0, 100,
             100, 100, 0, 0, 0, 0, 0, 98, 100, 0, 0, 100, 100, 0, 100, 0, 0, 0,
             100, 100, 0, 0, 100, 0, 94, 100, 72, 100, 100, 0, 0, 0, 100, 100,
             100, 100, 100, 100, 0, 0, 0, 0, 0, 0, 0, 94, 0, 100, 97, 100, 100,
             0, 100, 0, 0, 0, 0, 0, 81, 100, 97, 100, 81, 100, 0, 100, 100, 0,
             100, 100, 100, 100, 91, 0, 0, 0, 100, 0, 96, 100, 100, 99, 0, 0, 0,
             100, 0, 0, 0, 0, 0, 100, 0, 99, 100, 98, 0, 94, 0, 100, 0, 100,
             100, 0, 100, 100, 100, 0, 100, 0, 100, 67, 0, 91, 100, 100, 100,
             0, 0, 0, 0, 88, 0, 100, 0, 100, 0, 0]
            for x in range(len(hits)):
                assert(hits[x].pressure == expected_pressures[x])

        # Skip test cache since this test is calculating intensive events.
        run_analysis_test_on_replay(test, replay_list=get_raw_replays()["OCE_RLCS_7_CARS"],
                            answers=get_specific_answers()["CLEARS"],
                            calculate_intensive_events=True)



import pandas as pd
import math
from typing import List

from carball.generated.api.stats.events_pb2 import Hit
from carball.generated.api import game_pb2
from carball.json_parser.game import Game

class PressureAnalysis:
    def __init__(self, game: Game, proto_game: game_pb2, data_frame: pd.DataFrame):
        """Initialize a new instance of pressure analysis."""
        self.proto_game = proto_game
        self.game = game
        self.data_frame = data_frame

    def get_distance(self, position1: List[float], position2: List[float]) -> float:
        """Perform distance formula in any dimension between two positions."""
        if len(position1) != len(position2):
            return -1

        squared_distances = 0
        for dimension in range(len(position1)):
            squared_distances += (position1[dimension] - position2[dimension])**2
        return math.sqrt(squared_distances)

    def get_hit_pressure(self, hit: Hit):
        """Calculate pressure for the passed hit. Hit pressure determines how long it would have taken for the nearest opponent to hit the ball."""
        highest_pressure = 0
        teams = self.proto_game.teams
        opposing_team = None
        try:
            opposing_team = next(filter(lambda x: hit.player_id not in x.player_ids, teams))
        except StopIteration:
            hit.pressure = 0
            return
        opposing_team_ids = list(map(lambda x: x.id, opposing_team.player_ids))
        for player_id in opposing_team_ids:
            ball_row = self.data_frame.ball.iloc[hit.frame_number]
            player_name = None
            try:
                player_name = next(filter(lambda x: x.id.id == player_id, self.proto_game.players)).name
            except StopIteration:
                hit.pressure = 0
                break
            player_row = self.data_frame[player_name].iloc[hit.frame_number]

            # Get kinematics. Divide by 30 to consider velocity as uu/frame.
            ball_position = [ball_row.pos_x,ball_row.pos_y,ball_row.pos_z]
            player_position = [player_row.pos_x,player_row.pos_y,player_row.pos_z]
            player_velocity = [player_row.vel_x/30,player_row.vel_y/30,player_row.vel_z/30]

            original_distance = self.get_distance(ball_position, player_position)
            # Only consider defenders within 2000uu of hit.
            if original_distance < 2000:
                # Check the next 10 frames to see if the opposing player would have hit the ball.
                for i in range(100):
                    frame = i/10
                    estimated_player_position = [player_position[x]+frame*player_velocity[x] for x in range(len(ball_position))]
                    new_distance = self.get_distance(estimated_player_position,ball_position)
                    if new_distance > original_distance and new_distance > 500:
                        # Defender is not going towards ball, stop checking
                        break
                    if new_distance < 500:
                        highest_pressure = int(max(highest_pressure, (10-frame)*10))
                        break
        hit.pressure = highest_pressure

    def calculate_pressure_stats(self):
        """Will assign a pressure value to every hit in the game."""
        for hit in self.proto_game.game_stats.hits:
            self.get_hit_pressure(hit)



import logging

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .game import Game
    from .player import Player

logger = logging.getLogger(__name__)


class Goal:

    def __init__(self, goal_dict, game: 'Game'):
        # self.time = goal_dict["Time"]
        self.player_name = goal_dict["PlayerName"]
        self.player_team = goal_dict["PlayerTeam"]

        self.player = self.get_player(game)
        self.frame_number = goal_dict["frame"]
        logger.debug('Created Goal: %s', self)

    def __repr__(self):
        if self.player:
            return "Goal by %s on frame %s" % (self.player.name, self.frame_number)
        else:
            return "Goal by unknown player with name %s on frame %s" % (self.player_name, self.frame_number)

    def get_player(self, game: 'Game') -> 'Player':
        for player in game.players:
            if player.name == self.player_name:
                return player



import logging
from typing import List

logger = logging.getLogger(__name__)


class GameInfo:
    def __init__(self):
        self.server_id = None
        self.server_name = None
        self.match_guid = None
        self.game_mode = None
        self.mutator_index = None
        self.rumble_mutator = None

    def parse_game_info_actor(self, actor_data: dict, game_event_actor: dict, objects: List[str]):
        """
        Parses game actor
        :param actor_data: game replication info
        :param game_event_actor: soccar game event actor
        :param objects: object list from the decompiled replay
        :return: self
        """        
        # There is no GameServerID if you play alone
        self.server_id = ''
        if 'ProjectX.GRI_X:GameServerID' in actor_data:
            self.server_id = actor_data['ProjectX.GRI_X:GameServerID']
        self.server_name = actor_data['Engine.GameReplicationInfo:ServerName']
        
        # A custom lobby doesn't have a MatchGUID
        self.match_guid = actor_data.get('ProjectX.GRI_X:MatchGUID', '')
        self.playlist = actor_data['ProjectX.GRI_X:ReplicatedGamePlaylist']
        self.mutator_index = actor_data.get('ProjectX.GRI_X:ReplicatedGameMutatorIndex', 0)

        if 'TAGame.GameEvent_Soccar_TA:SubRulesArchetype' in game_event_actor:
            # Only used for rumble stats
            # TODO can this contain any other mutators?
            self.rumble_mutator = objects[game_event_actor['TAGame.GameEvent_Soccar_TA:SubRulesArchetype']['actor']]

        logger.info('Created game info from actor')
        return self



from carball.analysis.analysis_manager import AnalysisManager

from carball.tests.utils import run_analysis_test_on_replay, get_raw_replays


class Test_Kickoff():

    def test_0_kickoffs(self):
        def test(analysis: AnalysisManager):

            proto_game = analysis.get_protobuf_data()
            kickoffs = proto_game.game_stats.kickoffs
            assert (len(kickoffs) == 0)

            first_player = proto_game.players[0].stats
            kickoff_stats = first_player.kickoff_stats
            assert kickoff_stats.total_kickoffs == 0

        run_analysis_test_on_replay(test, get_raw_replays()["KICKOFF_NO_TOUCH"])

    def test_no_one_for_kickoffs(self):
        def test(analysis: AnalysisManager):

            proto_game = analysis.get_protobuf_data()
            kickoffs = proto_game.game_stats.kickoffs
            assert (len(kickoffs) == 1)

        run_analysis_test_on_replay(test, get_raw_replays()["NO_ONE_FOR_KICKOFF"])

    def test_no_one_for_weird_kickoffs(self):
        def test(analysis: AnalysisManager):

            proto_game = analysis.get_protobuf_data()
            kickoffs = proto_game.game_stats.kickoffs
            assert (len(kickoffs) == 3)

            first_player = proto_game.players[0].stats
            kickoff_stats = first_player.kickoff_stats
            assert kickoff_stats.total_kickoffs == 3

        run_analysis_test_on_replay(test, get_raw_replays()["WEIRD_KICKOFFS"])

    def test_3_kickoffs(self):
        def test(analysis: AnalysisManager):

            proto_game = analysis.get_protobuf_data()
            kickoffs = proto_game.game_stats.kickoffs
            assert (len(kickoffs) == 3)

            first_player = proto_game.players[0].stats
            kickoff_stats = first_player.kickoff_stats
            assert kickoff_stats.num_time_first_touch == 3
            assert kickoff_stats.num_time_go_to_ball == 3
            assert kickoff_stats.total_kickoffs == 3
            for i in range(len(kickoffs)):
                kickoff = proto_game.game_stats.kickoff_stats[i]
                player = kickoff.touch.players[0]
                start_left = player.start_position.pos_x > 0
                assert player.start_left == start_left
        run_analysis_test_on_replay(test, get_raw_replays()["3_KICKOFFS"])

    def test_last_kickoff_no_touch(self):

        def test(analysis: AnalysisManager):

            proto_game = analysis.get_protobuf_data()
            kickoffs = proto_game.game_stats.kickoffs
            assert (len(kickoffs) == 2)

            first_player = proto_game.players[0].stats
            kickoff_stats = first_player.kickoff_stats
            assert kickoff_stats.total_kickoffs == 2

        run_analysis_test_on_replay(test, get_raw_replays()["LAST_KICKOFF_NO_TOUCH"])

    def test_all_kickoffs(self):
        def test(analysis: AnalysisManager):

            proto_game = analysis.get_protobuf_data()
            kickoffs = proto_game.game_stats.kickoffs
            assert (len(kickoffs) == 5)
            first_player = proto_game.players[0].stats
            kickoff_stats = first_player.kickoff_stats
            assert kickoff_stats.num_time_first_touch == 0
            assert kickoff_stats.num_time_go_to_ball == 2
            assert kickoff_stats.num_time_boost == 1
            assert kickoff_stats.num_time_cheat == 1
            assert kickoff_stats.num_time_defend == 1
            assert kickoff_stats.total_kickoffs == 5

            second_player_stats = proto_game.players[1].stats.kickoff_stats
            assert second_player_stats.num_time_first_touch == 5
            assert second_player_stats.num_time_go_to_ball == 5
            for i in range(len(kickoffs)):
                kickoff = proto_game.game_stats.kickoff_stats[i]
                player = kickoff.touch.players[0]
                start_left = player.start_position.pos_x < 0
                assert player.start_left == start_left

        run_analysis_test_on_replay(test, get_raw_replays()["5_DIVERSE_KICKOFFS"])

    def test_all_kickoffs_2(self):
        def test(analysis: AnalysisManager):

            proto_game = analysis.get_protobuf_data()
            kickoffs = proto_game.game_stats.kickoffs
            assert (len(kickoffs) == 6)

            first_player = proto_game.players[0].stats
            kickoff_stats = first_player.kickoff_stats
            assert kickoff_stats.num_time_first_touch == 1
            assert kickoff_stats.num_time_go_to_ball == 1
            assert kickoff_stats.num_time_boost == 1
            assert kickoff_stats.num_time_cheat == 0
            assert kickoff_stats.num_time_defend == 1
            assert kickoff_stats.num_time_afk == 1
            assert kickoff_stats.total_kickoffs == 6
            assert 2 == (kickoff_stats.total_kickoffs -
                         (kickoff_stats.num_time_go_to_ball +
                          kickoff_stats.num_time_boost +
                          kickoff_stats.num_time_cheat +
                          kickoff_stats.num_time_defend +
                          kickoff_stats.num_time_afk))

            second_player_stats = proto_game.players[1].stats.kickoff_stats
            assert second_player_stats.num_time_first_touch == 5
            assert second_player_stats.num_time_go_to_ball == 6
            for i in range(len(kickoffs)):
                kickoff = proto_game.game_stats.kickoff_stats[i]
                player = kickoff.touch.players[0]
                start_left = player.start_position.pos_x < 0
                assert player.start_left == start_left

        run_analysis_test_on_replay(test, get_raw_replays()["6_DIVERSE_KICKOFFS"])



import unittest

from carball.analysis.analysis_manager import AnalysisManager

from carball.tests.utils import run_analysis_test_on_replay, get_raw_replays

from carball.generated.api.stats.rumble_pb2 import *


class Test_Rumble():

    def test_pre_item_goals(self, replay_cache):
        def test(analysis: AnalysisManager):
            proto_game = analysis.get_protobuf_data()

            assert proto_game.game_metadata.goals[0].extra_mode_info.pre_items
            assert not proto_game.game_metadata.goals[1].extra_mode_info.pre_items
            assert proto_game.game_metadata.goals[2].extra_mode_info.pre_items

        run_analysis_test_on_replay(test, get_raw_replays()["RUMBLE_PRE_ITEM_GOALS"], cache=replay_cache)

    def test_item_goals(self, replay_cache):
        def test(analysis: AnalysisManager):
            proto_game = analysis.get_protobuf_data()
            goals = proto_game.game_metadata.goals

            for i in range(5):
                assert goals[i].extra_mode_info.scored_with_item

            assert (goals[0].extra_mode_info.used_item == GRAVITY_WELL)
            assert (goals[1].extra_mode_info.used_item == BALL_GRAPPLING_HOOK)
            assert (goals[2].extra_mode_info.used_item == STRONG_HIT)
            assert (goals[3].extra_mode_info.used_item == BALL_VELCRO)
            assert (goals[4].extra_mode_info.used_item == BALL_LASSO)

            assert not goals[5].extra_mode_info.scored_with_item

        run_analysis_test_on_replay(test, get_raw_replays()["RUMBLE_ITEM_GOALS"], cache=replay_cache)

    def test_freeze_vs_spike(self, replay_cache):
        def test(analysis: AnalysisManager):
            proto_game = analysis.get_protobuf_data()

            assert (proto_game.game_stats.rumble_items[1].frame_number_use != -1)

            freeze_stats = next(filter(lambda x: x.item == BALL_FREEZE,
                                       proto_game.players[0].stats.rumble_stats.rumble_items))
            assert (freeze_stats.used == 1)

        run_analysis_test_on_replay(test, get_raw_replays()["RUMBLE_FREEZE_VS_SPIKE"], cache=replay_cache)

    def test_hold_time(self, replay_cache):
        assertions = unittest.TestCase('__init__')

        def test(analysis: AnalysisManager):
            proto_game = analysis.get_protobuf_data()

            spike_stats = next(filter(lambda x: x.item == BALL_VELCRO,
                                      proto_game.players[0].stats.rumble_stats.rumble_items))
            assertions.assertAlmostEqual(spike_stats.average_hold, 11.87916, 5)

            spike_stats = next(filter(lambda x: x.item == BALL_VELCRO,
                                      proto_game.teams[0].stats.rumble_stats.rumble_items))
            assertions.assertAlmostEqual(spike_stats.average_hold, 11.87916, 5)

        run_analysis_test_on_replay(test, get_raw_replays()["RUMBLE_HOLD_TIME"], cache=replay_cache)

    def test_item_count(self, replay_cache):
        def test(analysis: AnalysisManager):
            proto_game = analysis.get_protobuf_data()

            self.assert_rumble_item_counts(proto_game.players[0].stats.rumble_stats, [
                {'item': BALL_FREEZE, 'used': 1, 'unused': 0},
                {'item': BALL_GRAPPLING_HOOK, 'used': 1, 'unused': 0},
                {'item': BALL_LASSO, 'used': 1, 'unused': 0},
                {'item': BALL_SPRING, 'used': 2, 'unused': 0},
                {'item': BALL_VELCRO, 'used': 0, 'unused': 1},
                {'item': BOOST_OVERRIDE, 'used': 2, 'unused': 0},
                {'item': CAR_SPRING, 'used': 1, 'unused': 1},
                {'item': GRAVITY_WELL, 'used': 2, 'unused': 0},
                {'item': STRONG_HIT, 'used': 2, 'unused': 0},
                {'item': SWAPPER, 'used': 1, 'unused': 0},
                {'item': TORNADO, 'used': 0, 'unused': 0}
            ])

            self.assert_rumble_item_counts(proto_game.players[1].stats.rumble_stats, [
                {'item': BALL_FREEZE, 'used': 2, 'unused': 0},
                {'item': BALL_GRAPPLING_HOOK, 'used': 1, 'unused': 1},
                {'item': BALL_LASSO, 'used': 1, 'unused': 0},
                {'item': BALL_SPRING, 'used': 1, 'unused': 0},
                {'item': BALL_VELCRO, 'used': 0, 'unused': 1},
                {'item': BOOST_OVERRIDE, 'used': 2, 'unused': 0},
                {'item': CAR_SPRING, 'used': 0, 'unused': 1},
                {'item': GRAVITY_WELL, 'used': 2, 'unused': 0},
                {'item': STRONG_HIT, 'used': 1, 'unused': 0},
                {'item': SWAPPER, 'used': 2, 'unused': 0},
                {'item': TORNADO, 'used': 0, 'unused': 1}
            ])

            self.assert_rumble_item_counts(proto_game.players[2].stats.rumble_stats, [
                {'item': BALL_FREEZE, 'used': 1, 'unused': 0},
                {'item': BALL_GRAPPLING_HOOK, 'used': 0, 'unused': 0},
                {'item': BALL_LASSO, 'used': 2, 'unused': 0},
                {'item': BALL_SPRING, 'used': 1, 'unused': 0},
                {'item': BALL_VELCRO, 'used': 1, 'unused': 1},
                {'item': BOOST_OVERRIDE, 'used': 1, 'unused': 1},
                {'item': CAR_SPRING, 'used': 0, 'unused': 0},
                {'item': GRAVITY_WELL, 'used': 2, 'unused': 0},
                {'item': STRONG_HIT, 'used': 2, 'unused': 0},
                {'item': SWAPPER, 'used': 1, 'unused': 1},
                {'item': TORNADO, 'used': 1, 'unused': 0}
            ])

            self.assert_rumble_item_counts(proto_game.players[3].stats.rumble_stats, [
                {'item': BALL_FREEZE, 'used': 0, 'unused': 1},
                {'item': BALL_GRAPPLING_HOOK, 'used': 1, 'unused': 1},
                {'item': BALL_LASSO, 'used': 1, 'unused': 0},
                {'item': BALL_SPRING, 'used': 1, 'unused': 0},
                {'item': BALL_VELCRO, 'used': 1, 'unused': 0},
                {'item': BOOST_OVERRIDE, 'used': 2, 'unused': 0},
                {'item': CAR_SPRING, 'used': 2, 'unused': 0},
                {'item': GRAVITY_WELL, 'used': 1, 'unused': 0},
                {'item': STRONG_HIT, 'used': 1, 'unused': 0},
                {'item': SWAPPER, 'used': 2, 'unused': 0},
                {'item': TORNADO, 'used': 0, 'unused': 0}
            ])

            self.assert_rumble_item_counts(proto_game.players[4].stats.rumble_stats, [
                {'item': BALL_FREEZE, 'used': 2, 'unused': 0},
                {'item': BALL_GRAPPLING_HOOK, 'used': 2, 'unused': 0},
                {'item': BALL_LASSO, 'used': 2, 'unused': 0},
                {'item': BALL_SPRING, 'used': 0, 'unused': 0},
                {'item': BALL_VELCRO, 'used': 1, 'unused': 0},
                {'item': BOOST_OVERRIDE, 'used': 2, 'unused': 0},
                {'item': CAR_SPRING, 'used': 1, 'unused': 0},
                {'item': GRAVITY_WELL, 'used': 2, 'unused': 0},
                {'item': STRONG_HIT, 'used': 1, 'unused': 0},
                {'item': SWAPPER, 'used': 1, 'unused': 1},
                {'item': TORNADO, 'used': 0, 'unused': 0}
            ])

            self.assert_rumble_item_counts(proto_game.players[5].stats.rumble_stats, [
                {'item': BALL_FREEZE, 'used': 2, 'unused': 0},
                {'item': BALL_GRAPPLING_HOOK, 'used': 2, 'unused': 0},
                {'item': BALL_LASSO, 'used': 2, 'unused': 0},
                {'item': BALL_SPRING, 'used': 1, 'unused': 0},
                {'item': BALL_VELCRO, 'used': 1, 'unused': 0},
                {'item': BOOST_OVERRIDE, 'used': 1, 'unused': 0},
                {'item': CAR_SPRING, 'used': 2, 'unused': 0},
                {'item': GRAVITY_WELL, 'used': 0, 'unused': 0},
                {'item': STRONG_HIT, 'used': 2, 'unused': 0},
                {'item': SWAPPER, 'used': 1, 'unused': 1},
                {'item': TORNADO, 'used': 1, 'unused': 0}
            ])

            self.assert_rumble_item_counts(proto_game.teams[0].stats.rumble_stats, [
                {'item': BALL_FREEZE, 'used': 2, 'unused': 1},
                {'item': BALL_GRAPPLING_HOOK, 'used': 2, 'unused': 1},
                {'item': BALL_LASSO, 'used': 4, 'unused': 0},
                {'item': BALL_SPRING, 'used': 4, 'unused': 0},
                {'item': BALL_VELCRO, 'used': 2, 'unused': 2},
                {'item': BOOST_OVERRIDE, 'used': 5, 'unused': 1},
                {'item': CAR_SPRING, 'used': 3, 'unused': 1},
                {'item': GRAVITY_WELL, 'used': 5, 'unused': 0},
                {'item': STRONG_HIT, 'used': 5, 'unused': 0},
                {'item': SWAPPER, 'used': 4, 'unused': 1},
                {'item': TORNADO, 'used': 1, 'unused': 0}
            ])

            self.assert_rumble_item_counts(proto_game.teams[1].stats.rumble_stats, [
                {'item': BALL_FREEZE, 'used': 6, 'unused': 0},
                {'item': BALL_GRAPPLING_HOOK, 'used': 5, 'unused': 1},
                {'item': BALL_LASSO, 'used': 5, 'unused': 0},
                {'item': BALL_SPRING, 'used': 2, 'unused': 0},
                {'item': BALL_VELCRO, 'used': 2, 'unused': 1},
                {'item': BOOST_OVERRIDE, 'used': 5, 'unused': 0},
                {'item': CAR_SPRING, 'used': 3, 'unused': 1},
                {'item': GRAVITY_WELL, 'used': 4, 'unused': 0},
                {'item': STRONG_HIT, 'used': 4, 'unused': 0},
                {'item': SWAPPER, 'used': 4, 'unused': 2},
                {'item': TORNADO, 'used': 1, 'unused': 1}
            ])

        run_analysis_test_on_replay(test, get_raw_replays()["RUMBLE_FULL"], cache=replay_cache)

    def test_item_kickoff(self, replay_cache):
        def test(analysis: AnalysisManager):
            proto_game = analysis.get_protobuf_data()
            self.assert_rumble_item_counts(proto_game.players[0].stats.rumble_stats, [
                {'item': BALL_FREEZE, 'used': 0, 'unused': 0},
                {'item': BALL_GRAPPLING_HOOK, 'used': 0, 'unused': 0},
                {'item': BALL_LASSO, 'used': 0, 'unused': 0},
                {'item': BALL_SPRING, 'used': 2, 'unused': 0},
                {'item': BALL_VELCRO, 'used': 0, 'unused': 0},
                {'item': BOOST_OVERRIDE, 'used': 0, 'unused': 0},
                {'item': CAR_SPRING, 'used': 0, 'unused': 0},
                {'item': GRAVITY_WELL, 'used': 0, 'unused': 0},
                {'item': STRONG_HIT, 'used': 0, 'unused': 0},
                {'item': SWAPPER, 'used': 0, 'unused': 0},
                {'item': TORNADO, 'used': 0, 'unused': 0}
            ])
        run_analysis_test_on_replay(test, get_raw_replays()["RUMBLE_ITEM_KICKOFF"], cache=replay_cache)

    def assert_rumble_item_counts(self, rumble_stats_proto, expected):
        result_stats = list(map(proto_to_dict, rumble_stats_proto.rumble_items))
        case = unittest.TestCase('__init__')
        case.assertCountEqual(result_stats, expected)


def proto_to_dict(item_proto):
    return {
        'item': item_proto.item,
        'used': item_proto.used,
        'unused': item_proto.unused
    }



from .base import *


class GameEventHandler(BaseActorHandler):

    @classmethod
    def can_handle(cls, actor: dict) -> bool:
        if actor['ClassName'] is not None:
            return actor['ClassName'].startswith('TAGame.GameEvent_Soccar_TA') or actor['ClassName'].startswith('TAGame.GameEvent_GodBall_TA')
        else:
            return False

    def update(self, actor: dict, frame_number: int, time: float, delta: float) -> None:
        self.parser.soccar_game_event_actor = actor
        frame_data = {
            'time': time,
            'delta': delta,
            'seconds_remaining': actor.get('TAGame.GameEvent_Soccar_TA:SecondsRemaining', None),
            'replicated_seconds_remaining': actor.get('TAGame.GameEvent_TA:ReplicatedGameStateTimeRemaining', None),
            'is_overtime': actor.get('TAGame.GameEvent_Soccar_TA:bOverTime', None),
            'ball_has_been_hit': actor.get('TAGame.GameEvent_Soccar_TA:bBallHasBeenHit', None)
        }
        self.parser.frames_data[frame_number] = frame_data



import logging
from .base import *
from carball.json_parser.actor_parsing import CarActor

logger = logging.getLogger(__name__)


class CarHandler(BaseActorHandler):
    type_name = 'Archetypes.Car.Car_Default'

    def update(self, actor: dict, frame_number: int, time: float, delta: float) -> None:
        if 'Engine.Pawn:PlayerReplicationInfo' not in actor:
            return

        player_actor_id = actor['Engine.Pawn:PlayerReplicationInfo']['actor']
        if player_actor_id == -1:
            self.add_demo(actor, frame_number)
            return
        # assign car player links
        self.parser.player_car_ids[player_actor_id] = actor['Id']
        self.parser.car_player_ids[actor['Id']] = player_actor_id

        RBState = actor.get(REPLICATED_RB_STATE_KEY, {})
        # bDriving is missing?! TODO: Investigate bDriving in RBState
        # car_is_driving = RBState.get("rigid_body_state", {}).get("TAGame.Vehicle_TA:bDriving", False)
        car_is_sleeping = RBState.get('sleeping', True)
        # only collect data if car is driving and not sleeping
        if not car_is_sleeping:
            self.parser.current_car_ids_to_collect.append(actor['Id'])

            data_dict = CarActor.get_data_dict(actor)
            # save data from here
            self.parser.player_data[player_actor_id][frame_number].update(data_dict)

        # get demo data
        self.add_demo(actor, frame_number)

        if player_actor_id not in self.parser.car_dicts:
            self.parser.car_dicts[player_actor_id] = {'team_paint': {}}

        team_paint = actor['TAGame.Car_TA:TeamPaint']

        self.parser.car_dicts[player_actor_id]['team_paint'][team_paint['team']] = {
            'primary_color': team_paint['primary_color'],
            'accent_color': team_paint['accent_color'],
            'primary_finish': team_paint['primary_finish'],
            'accent_finish': team_paint['accent_finish']
        }

    def add_demo(self, actor, frame_number):
        if 'TAGame.Car_TA:ReplicatedDemolish' in actor:
            demo_data = actor['TAGame.Car_TA:ReplicatedDemolish']
            # add attacker and victim player ids
            attacker_car_id = demo_data['attacker']
            victim_car_id = demo_data['victim']
            if attacker_car_id != -1 and victim_car_id != -1 and attacker_car_id < 1e9 and victim_car_id < 1e9:
                # Filter out weird stuff where it's not a demo
                # frame 1 of 0732D41D4AF83D610AE2A988ACBC977A (rlcs season 4 eu)
                attacker_player_id = self.parser.car_player_ids[attacker_car_id]
                victim_player_id = self.parser.car_player_ids[victim_car_id]
                if attacker_player_id != -1 and victim_player_id != -1:
                    demo_data['attacker_player_id'] = attacker_player_id
                    demo_data['victim_player_id'] = victim_player_id
                    # add frame_number
                    demo_data['frame_number'] = frame_number
                    self.parser.demos_data.append(demo_data)
                    logger.debug('ReplicatedDemolish: Att: %s, Def: %s' %
                                 (attacker_player_id, victim_player_id))
                    logger.debug('RepDemo Names: Att: %s. Def: %s' %
                                 (self.parser.player_dicts[attacker_player_id]['name'],
                                  self.parser.player_dicts[victim_player_id]['name']))
                    actor.pop('TAGame.Car_TA:ReplicatedDemolish')



from carball.generated.api.game_pb2 import Game

from carball.generated.api.metadata.game_metadata_pb2 import Playlist

from carball.generated.api.metadata import game_metadata_pb2


def get_team_size_from_playlist(playlist: Playlist):
    """
    Returns number of teams in the playlist.
    5 if it is unknown otherwise it will return a valid value.
    :param playlist:
    :return:
    """
    if (playlist == game_metadata_pb2.RANKED_DOUBLES or
            playlist == game_metadata_pb2.UNRANKED_DOUBLES or
            playlist == game_metadata_pb2.RANKED_HOOPS or
            playlist == game_metadata_pb2.UNRANKED_HOOPS):
        return 2

    if playlist == game_metadata_pb2.RANKED_DUELS or playlist == game_metadata_pb2.UNRANKED_DUELS:
        return 1

    if playlist == game_metadata_pb2.UNRANKED_CHAOS:
        return 4

    if playlist == game_metadata_pb2.CUSTOM_LOBBY or playlist == game_metadata_pb2.OFFLINE_SPLITSCREEN:
        return 5

    return 3


def get_playlist_from_game(proto_game: Game):
    return proto_game.game_metadata.playlist


def get_team_size_from_game(proto_game: Game):
    return get_team_size_from_playlist(get_playlist_from_game(proto_game))



import pandas as pd
from carball.generated.api.stats.events_pb2 import Bump

from carball.generated.api.player_id_pb2 import PlayerId
from carball.json_parser.game import Game

from carball.generated.api import game_pb2


class BumpAnalysis:
    def __init__(self, game: Game, proto_game: game_pb2):
        self.proto_game = proto_game

    def get_bumps_from_game(self, data_frame: pd.DataFrame):
        self.create_bumps_from_demos(self.proto_game)

        self.analyze_bumps(data_frame)

    def create_bumps_from_demos(self, proto_game):
        for demo in proto_game.game_metadata.demos:
            self.add_bump(demo.frame_number, demo.victim_id, demo.attacker_id, True)

    def add_bump(self, frame: int, victim_id: PlayerId, attacker_id: PlayerId, is_demo: bool) -> Bump:
        bump = self.proto_game.game_stats.bumps.add()
        bump.frame_number = frame
        bump.attacker_id.id = attacker_id.id
        bump.victim_id.id = victim_id.id
        if is_demo:
            bump.is_demo = True

    def analyze_bumps(self, data_frame:pd.DataFrame):
        for bump in self.proto_game.game_stats.bumps:
            self.analyze_bump(bump, data_frame)


    def analyze_bump(self, bump: Bump, data_frame:pd.DataFrame):
        frame_number = bump.frame_number



import pandas as pd
import numpy as np

from carball.generated.api.stats.events_pb2 import FiftyFifty
from carball.generated.api.stats.events_pb2 import Hit
from carball.generated.api import game_pb2
from carball.json_parser.game import Game
from carball.analysis.constants.field_constants import HEIGHT_1_LIM, STANDARD_FIELD_LENGTH_HALF, MAX_CAR_SPEED

class FiftyAnalysis:
    def __init__(self, game: Game, proto_game: game_pb2, data_frame: pd.DataFrame):
        """Initialize a new instance of 50/50 analysis."""
        self.proto_game = proto_game
        self.game = game
        self.data_frame = data_frame
        self.fifty_threshold = 5

    def hits_are_consecutive(self, hit1: Hit, hit2: Hit):
        """Return True if hit2 occurs right after hit1 and are from different players."""
        withinThreshold = (hit2.frame_number - hit1.frame_number) < self.fifty_threshold
        opposingPlayers = hit1.player_id.id != hit2.player_id.id
        return withinThreshold and opposingPlayers

    def determine_fifty_fifty_winner(self, fifty: FiftyFifty, next_hit_index: int):
        """Iterate all 50/50s and determine the winner of each."""
        hits = self.proto_game.game_stats.hits
        hit = hits[next_hit_index]
        following_hit = hits[next_hit_index+1]
        last_hit_of_fifty = hits[next_hit_index-1]

        # If the last hit was a very useful hit, then it was won by the hitter.
        if last_hit_of_fifty.goal or last_hit_of_fifty.save or last_hit_of_fifty.clear or last_hit_of_fifty.assist or last_hit_of_fifty.pass_:
            fifty.winner.CopyFrom(last_hit_of_fifty.player_id)
            return

        # If the next hit starts a 50/50, then this hit was a neutral 50/50.
        if self.hits_are_consecutive(hit,following_hit):
            fifty.is_neutral = True
            return

        # Check for dunks.
        # If the first hit of the fifty is a kickoff, don't care about dunks.
        if hits[fifty.hits[0]].is_kickoff:
            fifty.is_neutral = True
            return

        last_hit_frame = last_hit_of_fifty.frame_number

        goal_pos_orange = [0, STANDARD_FIELD_LENGTH_HALF, HEIGHT_1_LIM/2]
        goal_pos_blue = [0, -1*STANDARD_FIELD_LENGTH_HALF, HEIGHT_1_LIM/2]

        # Get the ball position at the time of the hit.
        # Consider the vector from this position to each goal.
        ball_row = self.data_frame.ball.iloc[last_hit_frame+self.fifty_threshold]
        pos_vector_to_blue = [goal_pos_blue[0] - ball_row.pos_x, goal_pos_blue[1] - ball_row.pos_y, goal_pos_blue[2] - ball_row.pos_z]
        pos_vector_to_orange = [goal_pos_orange[0] - ball_row.pos_x, goal_pos_orange[1] - ball_row.pos_y, goal_pos_orange[2] - ball_row.pos_z]

        # Test the ball velocity slightly in the future so that it experiences no more acceleration from any hit.
        ball_row = self.data_frame.ball.iloc[last_hit_frame+self.fifty_threshold]
        ball_velocity = [ball_row.vel_x,ball_row.vel_y,ball_row.vel_z]

        # Find the component of velocity projected onto the vectors towards each goal.
        # If either dot product is significant, then the ball was speeding towards that goal.
        projection_to_blue_goal = np.dot(np.array(ball_velocity), np.array(pos_vector_to_blue))/np.linalg.norm(np.array(pos_vector_to_blue))
        projection_to_orange_goal = np.dot(np.array(ball_velocity), np.array(pos_vector_to_orange))/np.linalg.norm(np.array(pos_vector_to_orange))
        towards_blue_net = projection_to_blue_goal > MAX_CAR_SPEED and projection_to_orange_goal < -1*MAX_CAR_SPEED
        towards_orange_net = projection_to_orange_goal > MAX_CAR_SPEED and projection_to_blue_goal < -1*MAX_CAR_SPEED

        # If the last hit travelled for a while.
        last_hit_travel_time = hit.frame_number - last_hit_of_fifty.frame_number
        if last_hit_travel_time > self.fifty_threshold * 5:
            # Check if the ball was moving towards either net
            if towards_blue_net:
                # Find the last orange player to hit the ball towards the blue net. They are the winner.
                for i in range(len(fifty.hits)):
                    fifty_hit_index = fifty.hits[-1*i-1]
                    fifty_hit = hits[fifty_hit_index]
                    fifty_hit_player_id = fifty_hit.player_id
                    fifty_hit_player = next(filter(lambda x: x.id == fifty_hit_player_id,self.proto_game.players))
                    if fifty_hit_player and fifty_hit_player.is_orange:
                        fifty_hit.dunk = True
                        fifty.winner.CopyFrom(fifty_hit.player_id)
                        return
            if towards_orange_net:
                # Find the last blue player to hit the ball towards the orange net. They are the winner.
                for i in range(len(fifty.hits)):
                    fifty_hit_index = fifty.hits[-1*i-1]
                    fifty_hit=hits[fifty_hit_index]
                    fifty_hit_player_id = fifty_hit.player_id
                    fifty_hit_player = next(filter(lambda x: x.id == fifty_hit_player_id,self.proto_game.players))
                    if fifty_hit_player and not fifty_hit_player.is_orange:
                        fifty_hit.dunk = True
                        fifty.winner.CopyFrom(fifty_hit.player_id)
                        return

        # Nothing special happened after 50/50.
        fifty.is_neutral = True

    def calculate_fifty_fifty_stats(self):
        """Find all 50/50s in a match and assign their winners."""
        hits = self.proto_game.game_stats.hits
        # Check all hits for 50/50 potential.
        i = 0
        while i < len(hits)-1:
            hit = hits[i]
            lookahead = hits[i+1]
            hits_in_fifty = []
            players_involved = {}
            # If two hits are within threshold, we are starting a 50/50.
            if not self.hits_are_consecutive(hit,lookahead):
                # If no 50/50 found, go to next hit.
                i += 1
                continue
            else:
                # 50/50 detected.
                hits_in_fifty.append(i)
                hits_in_fifty.append(i+1)
                players_involved[hit.player_id.id] = hit.player_id
                players_involved[lookahead.player_id.id] = lookahead.player_id
            # Check next hit, break if that was the last hit.
            i += 1
            if i >= len(hits)-1:
                break
            hit = hits[i]
            lookahead = hits[i+1]
            # Start iterating hits that are all in 50/50.
            while self.hits_are_consecutive(hit, lookahead):
                hits_in_fifty.append(i+1)
                players_involved[lookahead.player_id.id] = lookahead.player_id
                # Break if that was the last hit.
                i += 1
                if i >= len(hits)-1:
                    break
                hit = hits[i]
                lookahead = hits[i+1]
            # Build 50/50 object.
            if len(hits_in_fifty) >= 1:
                fifty = self.proto_game.game_stats.fifty_fifties.add()
                fifty.hits.extend(hits_in_fifty)
                fifty.players.extend(list(players_involved.values()))
                fifty.starting_frame = hits[hits_in_fifty[0]].frame_number
                fifty.ending_frame = hits[hits_in_fifty[-1]].frame_number
                self.determine_fifty_fifty_winner(fifty, i+1)



OUT_OF_FRAME_COORDS = (10000, 10000)
MAP_X, MAP_Y, MAP_Z = 8200, 10280, 2050  # field dimensions
GOAL_X, GOAL_Y, GOAL_Z = 1792, 900, 640


MAP_X_LIMS = (-MAP_X / 2, MAP_X / 2)
MAP_Y_LIMS = (-MAP_Y / 2, MAP_Y / 2)
MAP_Z_LIMS = (0, MAP_Z)




import unittest

from carball.analysis.analysis_manager import AnalysisManager

from carball.tests.utils import run_analysis_test_on_replay, get_raw_replays
from carball.generated.api.stats.dropshot_pb2 import DAMAGED, DESTROYED


class Test_Dropshot():

    def test_single_damage(self, replay_cache):
        def test(analysis: AnalysisManager):
            assertions = unittest.TestCase('__init__')
            proto_game = analysis.get_protobuf_data()

            assert proto_game.players[0].stats.dropshot_stats.total_damage == 1
            assert proto_game.players[0].stats.dropshot_stats.damage_efficiency == 1.0

            assert proto_game.teams[0].stats.dropshot_stats.total_damage == 1
            assert proto_game.teams[0].stats.dropshot_stats.damage_efficiency == 1.0

            dropshot_ball_proto = proto_game.game_stats.ball_stats.extra_mode

            assert dropshot_ball_proto.dropshot_phase_stats[0].phase == 0
            assertions.assertAlmostEqual(dropshot_ball_proto.dropshot_phase_stats[0].average, 9.508282, 5)
            assertions.assertAlmostEqual(dropshot_ball_proto.dropshot_phase_stats[0].max, 9.508282, 5)
            assertions.assertAlmostEqual(dropshot_ball_proto.dropshot_phase_stats[0].total, 9.508282, 5)

            assert dropshot_ball_proto.dropshot_phase_stats[1].phase == 1
            assert dropshot_ball_proto.dropshot_phase_stats[1].average == 0
            assert dropshot_ball_proto.dropshot_phase_stats[1].max == 0
            assert dropshot_ball_proto.dropshot_phase_stats[1].total == 0

            assert dropshot_ball_proto.dropshot_phase_stats[2].phase == 2
            assert dropshot_ball_proto.dropshot_phase_stats[2].average == 0
            assert dropshot_ball_proto.dropshot_phase_stats[2].max == 0
            assert dropshot_ball_proto.dropshot_phase_stats[2].total == 0

            game_dropshot_stats = proto_game.game_stats.dropshot_stats

            assert len(game_dropshot_stats.damage_events) == 1
            assert len(game_dropshot_stats.damage_events[0].tiles) == 1
            assert game_dropshot_stats.damage_events[0].tiles[0].id == 101
            assert game_dropshot_stats.damage_events[0].tiles[0].state == DAMAGED

        run_analysis_test_on_replay(test, get_raw_replays()["DROPSHOT_SINGLE_DAMAGE"], cache=replay_cache)

    def test_double_damage(self, replay_cache):
        def test(analysis: AnalysisManager):
            proto_game = analysis.get_protobuf_data()
            assertions = unittest.TestCase('__init__')

            assert proto_game.players[0].stats.dropshot_stats.total_damage == 3
            assert proto_game.players[0].stats.dropshot_stats.damage_efficiency == 1.0

            assert proto_game.teams[1].stats.dropshot_stats.total_damage == 3
            assert proto_game.teams[1].stats.dropshot_stats.damage_efficiency == 1.0

            dropshot_ball_proto = proto_game.game_stats.ball_stats.extra_mode

            assert dropshot_ball_proto.dropshot_phase_stats[0].phase == 0
            assertions.assertAlmostEqual(dropshot_ball_proto.dropshot_phase_stats[0].average, 10.189136, 5)
            assertions.assertAlmostEqual(dropshot_ball_proto.dropshot_phase_stats[0].max, 10.189136, 5)
            assertions.assertAlmostEqual(dropshot_ball_proto.dropshot_phase_stats[0].total, 10.189136, 5)

            assert dropshot_ball_proto.dropshot_phase_stats[1].phase == 1
            assert dropshot_ball_proto.dropshot_phase_stats[1].average == 0
            assert dropshot_ball_proto.dropshot_phase_stats[1].max == 0
            assert dropshot_ball_proto.dropshot_phase_stats[1].total == 0

            assert dropshot_ball_proto.dropshot_phase_stats[2].phase == 2
            assert dropshot_ball_proto.dropshot_phase_stats[2].average == 0
            assert dropshot_ball_proto.dropshot_phase_stats[2].max == 0
            assert dropshot_ball_proto.dropshot_phase_stats[2].total == 0

            game_dropshot_stats = proto_game.game_stats.dropshot_stats

            assert len(game_dropshot_stats.damage_events) == 3
            assert game_dropshot_stats.damage_events[0].tiles[0].state == DAMAGED
            assert game_dropshot_stats.damage_events[1].tiles[0].state == DESTROYED
            assert len(game_dropshot_stats.tile_stats.damage_stats) == 2
            assert game_dropshot_stats.tile_stats.damage_stats[0].total_damage == 2

        run_analysis_test_on_replay(test, get_raw_replays()["DROPSHOT_DOUBLE_DAMAGE"], cache=replay_cache)

    def test_phase1_ball(self, replay_cache):
        def test(analysis: AnalysisManager):
            proto_game = analysis.get_protobuf_data()
            assertions = unittest.TestCase('__init__')

            dropshot_ball_proto = proto_game.game_stats.ball_stats.extra_mode

            assert dropshot_ball_proto.dropshot_phase_stats[0].phase == 0
            assertions.assertAlmostEqual(dropshot_ball_proto.dropshot_phase_stats[0].average, 3.72452275, 5)
            assertions.assertAlmostEqual(dropshot_ball_proto.dropshot_phase_stats[0].max, 6.9999535, 5)
            assertions.assertAlmostEqual(dropshot_ball_proto.dropshot_phase_stats[0].total, 7.4490455, 5)

            assert dropshot_ball_proto.dropshot_phase_stats[1].phase == 1
            assertions.assertAlmostEqual(dropshot_ball_proto.dropshot_phase_stats[1].average, 0.1, 5)
            assertions.assertAlmostEqual(dropshot_ball_proto.dropshot_phase_stats[1].max, 0.1, 5)
            assertions.assertAlmostEqual(dropshot_ball_proto.dropshot_phase_stats[1].total, 0.1, 5)

            assert dropshot_ball_proto.dropshot_phase_stats[2].phase == 2
            assert dropshot_ball_proto.dropshot_phase_stats[2].average == 0
            assert dropshot_ball_proto.dropshot_phase_stats[2].max == 0
            assert dropshot_ball_proto.dropshot_phase_stats[2].total == 0

            game_dropshot_stats = proto_game.game_stats.dropshot_stats

            assert len(game_dropshot_stats.damage_events[0].tiles) == 7

        run_analysis_test_on_replay(test, get_raw_replays()["DROPSHOT_PHASE1_BALL"], cache=replay_cache)

    def test_phase2_ball(self, replay_cache):
        def test(analysis: AnalysisManager):
            proto_game = analysis.get_protobuf_data()
            assertions = unittest.TestCase('__init__')

            dropshot_ball_proto = proto_game.game_stats.ball_stats.extra_mode

            assert dropshot_ball_proto.dropshot_phase_stats[0].phase == 0
            assertions.assertAlmostEqual(dropshot_ball_proto.dropshot_phase_stats[0].average, 7.12466465, 5)
            assertions.assertAlmostEqual(dropshot_ball_proto.dropshot_phase_stats[0].max, 12.1493613, 5)
            assertions.assertAlmostEqual(dropshot_ball_proto.dropshot_phase_stats[0].total, 14.2493293, 5)

            assert dropshot_ball_proto.dropshot_phase_stats[1].phase == 1
            assertions.assertAlmostEqual(dropshot_ball_proto.dropshot_phase_stats[1].average, 19.000713, 5)
            assertions.assertAlmostEqual(dropshot_ball_proto.dropshot_phase_stats[1].max, 19.000713, 5)
            assertions.assertAlmostEqual(dropshot_ball_proto.dropshot_phase_stats[1].total, 19.000713, 5)

            assert dropshot_ball_proto.dropshot_phase_stats[2].phase == 2
            assertions.assertAlmostEqual(dropshot_ball_proto.dropshot_phase_stats[2].average, 4.901173, 5)
            assertions.assertAlmostEqual(dropshot_ball_proto.dropshot_phase_stats[2].max, 4.901173, 5)
            assertions.assertAlmostEqual(dropshot_ball_proto.dropshot_phase_stats[2].total, 4.901173, 5)

            game_dropshot_stats = proto_game.game_stats.dropshot_stats

            assert len(game_dropshot_stats.damage_events[0].tiles) == 12

        run_analysis_test_on_replay(test, get_raw_replays()["DROPSHOT_PHASE2_BALL"], cache=replay_cache)

    def test_goal(self, replay_cache):
        def test(analysis: AnalysisManager):
            proto_game = analysis.get_protobuf_data()

            assert proto_game.game_metadata.goals[0].extra_mode_info.dropshot_tile.id == 89
            assert proto_game.game_metadata.goals[0].extra_mode_info.phase_1_tiles == 0
            assert proto_game.game_metadata.goals[0].extra_mode_info.phase_2_tiles == 1

        run_analysis_test_on_replay(test, get_raw_replays()["DROPSHOT_GOAL"], cache=replay_cache)



import logging
import time
from typing import Dict, List
from bisect import bisect_left

from ....generated.api import game_pb2
from ....generated.api.player_pb2 import Player
from ....generated.api.stats.events_pb2 import Hit
from carball.analysis.events.hit_detection.base_hit import BaseHit
from carball.analysis.simulator.ball_simulator import BallSimulator
from carball.analysis.simulator.map_constants import *
from carball.analysis.constants.field_constants import *


logger = logging.getLogger(__name__)


class SaltieHit:

    @staticmethod
    def get_distance_to_goal(data_frame: pd.DataFrame, hit: Hit, player_map: Dict[str, Player]):
        ball_data = BaseHit.get_ball_data(data_frame, hit)
        _goal_x = min(max(ball_data['pos_x'], GOAL_X / 2), -GOAL_X / 2)
        _goal_y = -MAP_Y / 2 if player_map[hit.player_id.id].is_orange else MAP_Y / 2

        displacement = ball_data[['pos_x', 'pos_y']].values - (_goal_x, _goal_y)
        distance = np.sqrt(np.square(displacement).sum())

        return distance

    @staticmethod
    def get_saltie_hits_from_game(proto_game: game_pb2.Game, hits: Dict[int, Hit],
                                  player_map: Dict[str, Player], data_frame: pd.DataFrame,
                                  kickoff_frames: pd.DataFrame) -> Dict[int, Hit]:
        hit_analytics_dict: Dict[int, Hit] = hits

        sorted_frames = sorted(hit_analytics_dict)

        start_time = time.time()
        SaltieHit.find_goal_hits(proto_game, kickoff_frames, sorted_frames, hit_analytics_dict)
        total_time = time.time() - start_time
        logger.debug('goal time: %s', total_time * 1000)

        SaltieHit.find_hit_stats(data_frame, player_map, sorted_frames, hit_analytics_dict)

        return hit_analytics_dict

    @staticmethod
    def find_goal_hits(proto_game: game_pb2.Game, kickoff_frames: pd.DataFrame,
                       sorted_frames: List[int], hit_analytics_dict: [int, Hit]):
        total_frames = len(sorted_frames)
        end_search = 0
        # find last hit by goalscorer for each goal frame
        for goal_number, goal in enumerate(proto_game.game_metadata.goals):
            goal_kickoff_frame = int(kickoff_frames[goal_number])
            last_goalscorer_saltie_hit = None
            # Only look for the frames that occur right before the goal.
            start_search = bisect_left(sorted_frames, goal.frame_number, end_search, total_frames - 1)
            for frame_index in range(start_search, end_search, -1):
                frame_number = sorted_frames[frame_index]
                saltie_hit = hit_analytics_dict[frame_number]
                if not (goal_kickoff_frame <= saltie_hit.frame_number <= goal.frame_number):
                    continue
                if saltie_hit.player_id.id == goal.player_id.id:
                    last_goalscorer_saltie_hit = saltie_hit
                    break
            end_search = start_search
            if last_goalscorer_saltie_hit is None:
                logger.warning("Could not find hit for goal: %s", goal)
            else:
                last_goalscorer_saltie_hit.goal = True
                # logger.debug("Found hit for goal on frame %s: %s", goal.frame_number, last_goalscorer_saltie_hit)

    @staticmethod
    def next_hit_stats(data_frame: pd.DataFrame, saltie_hit: Hit, next_saltie_hit: Hit,
                       player_map: Dict[str, Player], last_passing_hit: Hit):
        """
        finds stats that happen based off of the next hit.
        Passes, dribbles are found here, also candidates for assists.
        :param game:
        :param saltie_hit:
        :param next_saltie_hit:
        :param player_map:
        :param last_passing_hit:
        :return:
        """

        # distance the ball traveled
        displacement = (BaseHit.get_ball_data(data_frame, next_saltie_hit)[['pos_x', 'pos_y', 'pos_z']].values -
                        BaseHit.get_ball_data(data_frame, saltie_hit)[['pos_x', 'pos_y', 'pos_z']].values)
        saltie_hit.distance = np.sqrt(np.square(displacement).sum())

        # dribble detection
        if saltie_hit.player_id.id == next_saltie_hit.player_id.id:
            if not saltie_hit.dribble_continuation:
                saltie_hit.dribble = True
                next_saltie_hit.dribble_continuation = True
        else:
            last_passing_hit = None
            # passing detection
            if player_map[saltie_hit.player_id.id].is_orange == player_map[next_saltie_hit.player_id.id].is_orange:
                saltie_hit.pass_ = True
                next_saltie_hit.passed = True
                last_passing_hit = saltie_hit

        return last_passing_hit

    @staticmethod
    def get_shot(data_frame: pd.DataFrame, saltie_hit: Hit, player_map: Dict[str, Player]):
        """
        Finds shots using ball prediction.
        :param game:
        :param saltie_hit:
        :param player_map:
        """
        # find shots
        # TODO: Support non-standard maps? Raise warning/don't predict for non-standard maps?
        player = player_map[saltie_hit.player_id.id]
        ball_sim = BallSimulator(BaseHit.get_ball_data(data_frame, saltie_hit),
                                 player.is_orange)
        is_shot = ball_sim.get_is_shot()
        if is_shot:
            saltie_hit.shot = True
            # if saltie_hit.goal:
            #    logger.debug('Found shot for goal:')
        if saltie_hit.goal and not is_shot:
            logger.warning(f'Goal is not shot: frame {saltie_hit.frame_number} by {player.name}')

    @staticmethod
    def get_clear(data_frame: pd.DataFrame, saltie_hit: Hit, next_saltie_hit: Hit,  player_map: Dict[str, Player]):
        """
        Finds clears based on distance travelled and positions.
        :param game:
        :param saltie_hit:
        :param next_saltie_hit:
        :param player_map:
        """
        CLEAR_BUFFER = 400
        player =  player_map[saltie_hit.player_id.id]

        # get y-pos of ball to determine if the hit occurs in a player's defending third
        y_pos = saltie_hit.ball_data.pos_y
        defending_on_orange = (player.is_orange and y_pos > (STANDARD_FIELD_LENGTH_HALF/3 + CLEAR_BUFFER))
        defending_on_blue = (not player.is_orange and y_pos < ((-1)*STANDARD_FIELD_LENGTH_HALF/3 - CLEAR_BUFFER))

        # make sure the player is in their own defending third
        if not (defending_on_orange or defending_on_blue):
            return

        # determine if hit passed the buffer to be considered a clear
        if next_saltie_hit is not None:

            # check edge case where an own goal moves the ball to the middle
            goals_at_hit = data_frame.game.loc[saltie_hit.frame_number].goal_number
            goals_at_next_hit = data_frame.game.loc[next_saltie_hit.frame_number].goal_number
            if goals_at_hit != goals_at_next_hit:
                # lol they own-goaled
                return

            # find next hit, determine if this hit went far enough
            next_y = next_saltie_hit.ball_data.pos_y
            orange_reached_neutral_third = (player.is_orange and next_y < (STANDARD_FIELD_LENGTH_HALF/3 - CLEAR_BUFFER))
            blue_reached_neutral_third = (not player.is_orange and next_y > ((-1)*STANDARD_FIELD_LENGTH_HALF/3 + CLEAR_BUFFER))
            if orange_reached_neutral_third or blue_reached_neutral_third:
                saltie_hit.clear = True

        else:
            # a big hit to end the game should also count as a clear
            distance = saltie_hit.distance
            if distance > CLEAR_BUFFER:
                saltie_hit.clear = True

    @staticmethod
    def find_hit_stats(data_frame: pd.DataFrame, player_map: Dict[str, Player],
                       sorted_frames, hit_analytics_dict: Dict[int, Hit]):
        """
        Finds stats for all hits.
        :param data_frame:
        :param player_map:
        :param sorted_frames:
        :param hit_analytics_dict:
        """

        last_passing_hit = None
        total_stat_time = 0
        total_next_hit_time = 0
        total_simulation_time = 0
        for hit_number in range(len(sorted_frames)):
            start_time = time.time()
            hit_frame_number = sorted_frames[hit_number]
            saltie_hit: Hit = hit_analytics_dict[hit_frame_number]

            saltie_hit_goal_number = saltie_hit.goal_number
            # previous hit
            try:
                previous_hit_frame_number = sorted_frames[hit_number - 1]
                previous_saltie_hit = hit_analytics_dict[previous_hit_frame_number]
                if previous_saltie_hit.goal_number == saltie_hit_goal_number:
                    previous_saltie_hit = hit_analytics_dict[previous_hit_frame_number]
                    saltie_hit.previous_hit_frame_number = previous_saltie_hit.frame_number
            except IndexError:
                pass

            # next hit
            next_saltie_hit = None
            try:
                next_hit_frame_number = sorted_frames[hit_number + 1]
                next_saltie_hit = hit_analytics_dict[next_hit_frame_number]
                if next_saltie_hit.goal_number == saltie_hit_goal_number:
                    next_saltie_hit = hit_analytics_dict[next_hit_frame_number]
                    saltie_hit.next_hit_frame_number = next_saltie_hit.frame_number
            except IndexError:
                pass

            next_hit_time = time.time()
            total_next_hit_time += next_hit_time - start_time

            # aerials
            if saltie_hit.ball_data.pos_z >= 400.0:
                saltie_hit.aerial = True

            # assist calculation
            if saltie_hit.goal and last_passing_hit is not None:
                saltie_hit.assisted = True
                last_passing_hit.assist = True
                logger.debug('Found assist (%s) for goal', last_passing_hit.frame_number)

            # hit distance
            if next_saltie_hit:
                last_passing_hit = SaltieHit.next_hit_stats(data_frame, saltie_hit, next_saltie_hit,
                                                            player_map, last_passing_hit)
            elif saltie_hit.goal:
                saltie_hit.distance = 0

            stat_time = time.time()
            total_stat_time += stat_time - next_hit_time

            saltie_hit.distance_to_goal = SaltieHit.get_distance_to_goal(data_frame, saltie_hit, player_map)

            SaltieHit.get_shot(data_frame, saltie_hit, player_map)
            SaltieHit.get_clear(data_frame, saltie_hit, next_saltie_hit, player_map)

            simulation_time = time.time()
            total_simulation_time += simulation_time - stat_time

            try:
                if player_map[previous_saltie_hit.player_id.id].is_orange != player_map[saltie_hit.player_id.id].is_orange:
                    if previous_saltie_hit.shot and not previous_saltie_hit.goal:
                        saltie_hit.save = True
            except NameError:
                pass

        logger.debug('next time: %s', total_next_hit_time * 1000)
        logger.debug('stat time: %s', total_stat_time * 1000)
        logger.debug('sim  time: %s', total_simulation_time * 1000)



import argparse
import carball
import logging
import gzip


def main(program_args=None):
    parser = argparse.ArgumentParser(description='Rocket League replay parsing and analysis.')
    parser.add_argument('-i', '--input', type=str, required=True,
                        help='Path to replay file that will be analyzed. Carball expects a raw replay file unless '
                             '--skip-decompile is provided.')
    parser.add_argument('--proto', type=str, required=False,
                        help='The result of the analysis will be saved to this file in protocol buffers format.')
    parser.add_argument('--json', type=str, required=False,
                        help='The result of the analysis will be saved to this file in json file format.')
    parser.add_argument('--gzip', type=str, required=False,
                        help='The pandas data frame containing the replay frames will be saved to this file in a '
                             'compressed gzip format.')
    parser.add_argument('-v', '--verbose', action='count', default=0,
                        help='Set the logging level to INFO. To set the logging level to DEBUG use -vv.')
    parser.add_argument('-s', '--silent', action='store_true', default=False,
                        help='Disable logging altogether.')
    parser.add_argument('-dr', '--dry-run', action='store_true', default=False,
                        help='Explicitly notifying that there is not going to be any output to be saved')
    if program_args is not None:
        args = parser.parse_args(program_args)
    else:
        args = parser.parse_args()

    if not args.proto and not args.json and not args.gzip and not args.dry_run:
        parser.error('at least one of the following arguments are required: --proto, --json, --gzip')

    log_level = logging.WARNING

    if args.verbose == 1:
        log_level = logging.INFO
    elif args.verbose >= 2:
        log_level = logging.DEBUG

    if args.silent:
        logging.basicConfig(handlers=[logging.NullHandler()])
    else:
        logging.basicConfig(handlers=[logging.StreamHandler()], level=log_level)

    manager = carball.analyze_replay_file(args.input)

    if args.proto:
        with open(args.proto, 'wb') as f:
            manager.write_proto_out_to_file(f)
    if args.json:
        with open(args.json, 'w') as f:
            manager.write_json_out_to_file(f)
    if args.gzip:
        with gzip.open(args.gzip, 'wb') as f:
            manager.write_pandas_out_to_file(f)


if __name__ == '__main__':
    main()



import unittest
from carball.analysis.analysis_manager import AnalysisManager
from carball.tests.utils import run_analysis_test_on_replay, get_raw_replays


class Test_CameraSettings:

    def test_camera_settings(self, replay_cache):
        assertions = unittest.TestCase('__init__')

        def test(analysis: AnalysisManager):
            proto_game = analysis.get_protobuf_data()
            camera_settings = proto_game.players[0].camera_settings

            assertions.assertAlmostEqual(camera_settings.distance, 330.0, 5)
            assertions.assertAlmostEqual(camera_settings.field_of_view, 100.0, 5)
            assertions.assertAlmostEqual(camera_settings.height, 100.0, 5)
            assertions.assertAlmostEqual(camera_settings.pitch, -4.0, 5)
            assertions.assertAlmostEqual(camera_settings.stiffness, 0.45, 5)
            assertions.assertAlmostEqual(camera_settings.swivel_speed, 4.2, 5)
            assertions.assertAlmostEqual(camera_settings.transition_speed, 1.0, 5)

        run_analysis_test_on_replay(test, get_raw_replays()["DROPSHOT_GOAL"], cache=replay_cache)



import logging
import os
from boxcars_py import parse_replay

from carball.analysis.analysis_manager import AnalysisManager
from carball.controls.controls import ControlsCreator
from carball.extras.per_goal_analysis import PerGoalAnalysis
from carball.json_parser.game import Game
from carball.json_parser.sanity_check.sanity_check import SanityChecker

BASE_DIR = os.path.dirname(__file__)


def decompile_replay(replay_path):
    """
    Takes a path to the replay and outputs the json of that replay.

    :param replay_path: Path to a specific replay.
    :return: The object created from boxcars.
    """
    with open(replay_path, 'rb') as f:
        buf = f.read()
    return parse_replay(buf)


def analyze_replay_file(replay_path: str, controls: ControlsCreator = None,
                        sanity_check: SanityChecker = None, analysis_per_goal=False,
                        logging_level=logging.NOTSET,
                        calculate_intensive_events: bool = False,
                        clean: bool = True):
    """
    Decompile and analyze a replay file.

    :param replay_path: Path to replay file
    :param controls: Generate controls from the replay using our best guesses (ALPHA)
    :param sanity_check: Run sanity check to make sure we analyzed correctly (BETA)
    :param analysis_per_goal: Runs the analysis per a goal instead of the replay as a whole
    :param force_full_analysis: If True full analysis will be performed even if checks say it should not.
    :param logging_level: Sets the logging level globally across carball
    :param calculate_intensive_events: Indicates if expensive calculations should run to include additional stats.
    :param clean: Indicates if useless/invalid data should be found and removed.
    :return: AnalysisManager of game with analysis.
    """

    if logging_level != logging.NOTSET:
        logging.getLogger('carball').setLevel(logging_level)

    _json = decompile_replay(replay_path)
    game = Game()
    game.initialize(loaded_json=_json)
    # get_controls(game)  # TODO: enable and optimise.
    if sanity_check is not None:
        sanity_check.check_game(game)
    if analysis_per_goal:
        analysis = PerGoalAnalysis(game)
    else:
        analysis = AnalysisManager(game)
    analysis.create_analysis(calculate_intensive_events=calculate_intensive_events, clean=clean)

    if controls is not None:
        controls.get_controls(game)

    return analysis


if __name__ == '__main__':
    from carball.tests.analysis_test import __test_replays

    __test_replays(BASE_DIR)



import logging

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .player import Player

logger = logging.getLogger(__name__)


class Team:
    def __init__(self):
        self.is_orange = None
        self.name = None
        self.score = None
        self.actor_id = None

        self.players = set()

    def __repr__(self):
        if self.is_orange is not None:
            team_colour = 'Orange' if self.is_orange else 'Blue'
        else:
            team_colour = 'None'
        if self.name:
            return '%s: %s (%s)' % (self.__class__.__name__, self.name, team_colour)
        else:
            return '%s: %s' % (self.__class__.__name__, team_colour)

    def parse_team_data(self, team_data: dict):
        self.name = team_data.get("TAGame.Team_TA:CustomTeamName", None)
        self.is_orange = team_data['TypeName'].endswith('1')
        self.score = team_data.get("Engine.TeamInfo:Score", 0)  # RLCS S4 05E1F21043CCFCC442DCEC8B544A768F
        self.actor_id = team_data["Id"]

        logger.info('Created Team from stats: %s' % self)
        return self

    def add_player(self, player: 'Player'):
        self.players.add(player)
        player.team = self



import os
import shutil
from tempfile import mkstemp

from .create_proto import get_file_list, get_dir

import_statement = 'import '


def analyze_file(deepness, file_path, top_level_import):
    replace_map = dict()
    fh, abs_path = mkstemp()
    modified = False
    with os.fdopen(fh, 'w') as new_file:
        with open(file_path, 'r') as old_file:
            lines = old_file.readlines()
            for i in range(len(lines)):
                line = lines[i]
                extra_cases = '.' in line and top_level_import in line and not (import_statement + '.') in line
                if line.startswith(import_statement) and extra_cases:
                    cut_line = line[len(import_statement):].rstrip()
                    ending_string = cut_line[cut_line.rfind('.') + 1:]
                    replace_map[cut_line] = ending_string
                    line = 'from ' + '.' * deepness + cut_line[:cut_line.rfind(ending_string) - 1] + ' import ' + ending_string + '\n'
                    modified = True
                else:
                    for key in replace_map:
                        if key in line:
                            modified = True
                            line = line.replace(key, replace_map[key])
                new_file.write(line)
    if modified:
        os.remove(file_path)
        shutil.move(abs_path, file_path)
        print(str(len(replace_map)), 'imports')
    else:
        print('not modified')


def add_inits(top_level_dir):
    proto_dir = [x[0] for x in os.walk(get_dir()) if top_level_dir in x[0] and '__pycache__' not in x[0]]
    for directory in proto_dir:
        with open(os.path.join(directory, '__init__.py'), 'w') as file:
            print('creating', file.name)


def convert_to_relative_imports(top_level_dir='generated', exclude_dir=None, top_level_import="api"):
    print('###FIXING IMPORTS###')
    file_list = get_file_list(top_level_dir, exclude_dir)
    for file in file_list:
        print('fixing file: ', file[1], end='\t')
        analyze_file(file[0], file[1], top_level_import)
    add_inits(top_level_dir)


# prevent_leaks("carball", "generated", "carball")
if __name__ == '__main__':
    convert_to_relative_imports()

"""
Hi All  This file converts protobuf imports to relative imports so instead of
import api.player_pb2
it does
import ...api.player_pb2

The reason for this is described here:
https://github.com/protocolbuffers/protobuf/issues/1491

Basically google does not allow relative imports from their generated protobuf code.
So we do it so that our genreated file can go in a folder we dont care about

"""
